from crownstone_core.packets.assetFilter.InputDescriptionPackets import *

class AssetIdSourceBuilder:
    def __init__(self):
        self.inFormat: InputDescriptionPacket = None

    def build(self) -> InputDescriptionPacket:
        return self.inFormat

    def basedOnMac(self):
        """
        Base an asset ID on its MAC address.
        Use this if all assets (that pass the filter) have a static, and unique MAC address.
        """
        self.inFormat = InputDescriptionMacAddress()

    def basedOnName(self, complete: bool = True):
        """
        Base an asset ID on its name.
        Use this if all assets (that pass the filter) have a static, and unique name.

        :param complete:  Whether to look for the complete or shortened name.
        """
        adType = 0x09 if complete else 0x08
        self.inFormat = InputDescriptionFullAdData(adType)

    def basedOnManufacturerData(self):
        """
        Base an asset ID on the manufacturer data.
        Use this if all assets (that pass the filter) have static, and unique manufacturer data.
        """
        self.inFormat = InputDescriptionFullAdData(0xFF)

    def basedOnAdData(self, adType: int, bitmask: int = None):
        """
        Base an asset ID on AD data.
        :param adType:  The 8 bit GAP number.
                        See "Generic Access Profile" on https://www.bluetooth.com/specifications/assigned-numbers/
        :param bitmask: A 32 bits mask where the Nth bit represents the Nth byte in the AD data.
                        The data that is used as input for the filter, is a concatenation of all bytes that have their
                        associated bit set.
                        Example: if the AD data is: [10, 11, 12, 13, 14] and the bitmask is 22 (0000...00010110), then
                        the data that the filter uses is [11, 12, 14], and matched against each given asset.
        """
        if bitmask is None:
            self.inFormat = InputDescriptionFullAdData(adType)
        else:
            self.inFormat = InputDescriptionMaskedAdData(adType, bitmask)


from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter


class MicroappHeaderPacket(BasePacket):
	def __init__(self, appIndex: int=0, protocol: int=0):
		self.protocol = protocol
		self.appIndex = appIndex

	def _deserialize(self, reader: BufferReader):
		self.protocol = reader.getUInt8()
		self.appIndex = reader.getUInt8()

	def _serialize(self, writer: BufferWriter):
		writer.putUInt8(self.protocol)
		writer.putUInt8(self.appIndex)

	def __str__(self):
		return f"MicroappHeaderPacket(" \
		       f"protocol={self.protocol}, " \
		       f"appIndex={self.appIndex})"



from crownstone_core.packets.assetFilter.CuckooFilterPackets import CuckooFilterData
from crownstone_core.util.CRC import crc16ccitt, djb2_hash
from crownstone_core.util.Conversion import Conversion

from crownstone_core.util.Randomgenerator import RandomGenerator

# nice nonsensical value which will never be in uint16 range to mark an empty spot.
EMPTY = -1e8

class CuckooFilter:
    """
    Cuckoo filter implementation, currently supporting only 16 bit fingerprints.

    Byte arrays, 'keys', can be added to a cuckoo filter. They are stored as hashed, 'fingerprints',
    in an array. The array is organized in 'buckets' of fixed size. The fingerprint for a key is
    stored in one of two buckets, determined by a hash of the key. (The hash for fingerprints and
    the hash for buckets must be 'independent'.)

    Expected types for arguments and return values:
    - ByteArrayType = List[uint8]
    - FingerprintType = uint16
    - IndexType = uint8
    - FingerprintArrayType = List[FingerprintType]
    """
    max_kick_attempts = int(100)

    class ExtendedFingerprint:
        def __init__(self, fingerprint, bucketA, bucketB):
            self.fingerprint = fingerprint
            self.bucketA = bucketA
            self.bucketB = bucketB

        def __str__(self):
            return f"CuckooFilter.ExtendedFingerprint({self.fingerprint:#0{6}x},{self.bucketA:#0{4}x},{self.bucketB:#0{4}x})"

        def __eq__(self, other):
            """ Buckets are allowed to be reversed. Fingerprints must be equal. """
            return self.fingerprint == other.fingerprint and (
                   (self.bucketA == other.bucketA and self.bucketB == other.bucketB) or
                   (self.bucketB == other.bucketA and self.bucketA == other.bucketB)
            )

    def getExtendedFingerprint(self, key):
        """
        key: ByteArrayType
        returns ExtendedFingerprint
        """
        fingerprint = self.hash_to_fingerprint(key)
        buckethash = self.hash_to_bucket(key)

        return CuckooFilter.ExtendedFingerprint(
            fingerprint,
            buckethash % self.bucketCount(),
            (buckethash ^ fingerprint) % self.bucketCount())

    def getExtendedFingerprintFromFingerprintAndBucket(self, fingerprint, bucket_index):
        """
        fingerprint: FingerprintType
        bucket_index: IndexType
        returns: ExtendedFingerprint
        """
        bucket_a = (bucket_index % self.bucketCount()) & 0xff
        bucket_b =((bucket_index ^ fingerprint) % self.bucketCount()) & 0xff
        return CuckooFilter.ExtendedFingerprint (fingerprint, bucket_a, bucket_b)

    # -------------------------------------------------------------
    # Run time variables
    # -------------------------------------------------------------

    def __init__(self, bucketCountLog2, nests_per_bucket):
        """
        bucketCountLog2: IndexType. Number of buckets allocated is 2**bucketCountLog2.
        nests_per_bucket: IndexType
        returns: n/a
        """
        self.bucketCountLog2 = int(bucketCountLog2)
        self.nests_per_bucket = int(nests_per_bucket)
        self.victim = CuckooFilter.ExtendedFingerprint(0,0,0)
        self.bucket_array = []

        self.clear()

    # -------------------------------------------------------------
    # ----- Private methods -----
    # -------------------------------------------------------------

    def getData(self) -> CuckooFilterData:
        data = CuckooFilterData()
        data.bucketCountLog2 = self.bucketCountLog2
        data.nestsPerBucket = self.nests_per_bucket
        data.victim.fingerprint = self.victim.fingerprint
        data.victim.bucketA = self.victim.bucketA
        data.victim.bucketB = self.victim.bucketB
        data.bucketArray = self._cleanBucketArray(0)
        return data

    def serialize(self):
        return self.getData().serialize()

    def filterhash(self):
        """
        Flatten the filter settings and the uint16 array of fingerprints
        to a uint8 array in little endian and return a crc16 of the result.

        Must match firmware.

        returns: FingerprintType
        """
        return crc16ccitt(self.serialize())

    def hash_to_fingerprint(self, key):
        """
        key: ByteArrayType
        returns: FingerprintType
        """
        return crc16ccitt(key)

    def hash_to_bucket(self, key):
        """
        key: ByteArrayType
        returns: FingerprintType
        """
        return djb2_hash(key)

    def lookup_fingerprint(self, bucket_number, finger_index):
        """
        bucket_number: IndexType
        finger_index: IndexType
        returns: FingerprintType
        """
        return self.bucket_array[self.lookup_fingerprint_index(bucket_number, finger_index)]

    def lookup_fingerprint_index(self, bucket_number, finger_index):
        """
        bucket_number: IndexType
        finger_index: IndexType
        returns: int
        """
        return (bucket_number * self.nests_per_bucket) + finger_index

    def add_fingerprint_to_bucket (self, fingerprint, bucket_number):
        """
        fignerprint: FingerprintType
        bucket_number: IndexType
        returns: bool
        """
        for ii in range(self.nests_per_bucket):
            fingerprint_index = self.lookup_fingerprint_index(bucket_number, ii)
            if self.bucket_array[fingerprint_index] == EMPTY:
                self.bucket_array[fingerprint_index] = fingerprint
                return True
        return False

    def remove_fingerprint_from_bucket (self, fingerprint, bucket_number):
        """
        fignerprint: FingerprintType
        bucket_number: IndexType
        returns: bool
        """
        for ii in range(self.nests_per_bucket):
            candidate = self.lookup_fingerprint_index(bucket_number, ii) # candidate_fingerprint_for_removal_in_array_index

            if self.bucket_array[candidate] == fingerprint:
                self.bucket_array[candidate] = EMPTY
                # to keep the bucket front loaded, move the last non-zero
                # fingerprint behind ii into the slot.
                for jj in reversed(range(ii + 1, self.nests_per_bucket)):
                    last = self.lookup_fingerprint_index(bucket_number, jj) # last_fingerprint_in_bucket

                    if self.bucket_array[last] != EMPTY:
                        self.bucket_array[candidate] = self.bucket_array[last]
                        self.bucket_array[last] = EMPTY
                        break
                return True
        return False

    # -------------------------------------------------------------
    def moveExtendedFingerprint(self, entry_to_insert):
        """
        entry_to_insert: ExtendedFingerprint
        returns: bool
        """
        # seeding with a hash for this filter guarantees exact same sequence of
        # random integers used for moving fingerprints in the filter on every crownstone.
        seed = self.filterhash()
        rand = RandomGenerator(seed)

        for attempts_left in range(CuckooFilter.max_kick_attempts):
            # try to add to bucket A
            if self.add_fingerprint_to_bucket(entry_to_insert.fingerprint, entry_to_insert.bucketA):
                return True

            # try to add to bucket B
            if self.add_fingerprint_to_bucket(entry_to_insert.fingerprint, entry_to_insert.bucketB):
                return True

            # no success, time to kick a fingerprint from one of our buckets

            # determine which bucket to kick from
            kick_A = rand() % 2
            kicked_item_bucket =  entry_to_insert.bucketA if kick_A else entry_to_insert.bucketB

            # and which fingerprint index
            kicked_item_index = rand() % self.nests_per_bucket

            # swap entry to insert and the randomly chosen ('kicked') item
            kicked_item_fingerprint_index = self.lookup_fingerprint_index(kicked_item_bucket, kicked_item_index)
            kicked_item_fingerprint_value = self.bucket_array[kicked_item_fingerprint_index]

            self.bucket_array[kicked_item_fingerprint_index] = entry_to_insert.fingerprint

            entry_to_insert = self.getExtendedFingerprintFromFingerprintAndBucket(
                kicked_item_fingerprint_value, kicked_item_bucket)

            # next iteration will try to re-insert the footprint previously at (h,i).

        # iteration ended: failed to re-place the last kicked entry into the buffer after max attempts.
        self.victim = entry_to_insert

        return False

    def addExtendedFingerprint(self, extended_finger):
        """
        extended_finger: ExtendedFingerprint
        returns: bool
        """
        if self.containsExtendedFingerprint(extended_finger):
            return True

        if self.victim.fingerprint != 0: # already full.
            return False

        return self.moveExtendedFingerprint(extended_finger)

    def removeExtendedFingerprint(self, extended_finger):
        """
        extended_finger: ExtendedFingerprint
        returns: bool
        """
        if self.remove_fingerprint_from_bucket(extended_finger.fingerprint, extended_finger.bucketA) or \
                self.remove_fingerprint_from_bucket(extended_finger.fingerprint, extended_finger.bucketB):
            # short ciruits nicely:
            #    tries bucketA,
            #    on fail try B,
            #    if either succes, fix victim.
            if self.victim.fingerprint !=  0:
                if self.addExtendedFingerprint(self.victim):
                    self.victim = CuckooFilter.ExtendedFingerprint(0,0,0)
            return True
        return False

    def containsExtendedFingerprint(self, extended_finger):
        """
        extended_finger: ExtendedFingerprint
        returns: bool
        """
        # (loops are split to improve cache hit rate)
        # search bucketA
        for ii in range(self.nests_per_bucket):
            if extended_finger.fingerprint == self.lookup_fingerprint(extended_finger.bucketA, ii):
                return True
        # search bucketB
        for ii in range(self.nests_per_bucket):
            if extended_finger.fingerprint == self.lookup_fingerprint(extended_finger.bucketB, ii):
                return True

        return False

    # -------------------------------------------------------------

    def addFingerprintType(self, fingerprint, bucket_index):
        """
        fingerprint: FingerprintType
        returns: bool
        """
        return self.addExtendedFingerprint(
            self.getExtendedFingerprintFromFingerprintAndBucket(fingerprint, bucket_index))


    def removeFingerprintType(self, fingerprint, bucket_index):
        """
        fingerprint: FingerprintType
        bucket_index: Indextype
        returns: bool
        """
        return self.removeExtendedFingerprint(
            self.getExtendedFingerprintFromFingerprintAndBucket(fingerprint, bucket_index))


    def containsFingerprintType(self, fingerprint, bucket_index):
        """
        fingerprint: FingerprintType
        bucket_index: IndexType
        returns: bool
        """
        return self.containsExtendedFingerprint(
            self.getExtendedFingerprintFromFingerprintAndBucket(fingerprint, bucket_index))

    # -------------------------------------------------------------

    def add(self, key):
        """
        key: ByteArrayType
        returns: bool
        """
        return self.addExtendedFingerprint(self.getExtendedFingerprint(key))

    def remove(self, key):
        """
        key: ByteArrayType
        returns: bool
        """
        return self.removeExtendedFingerprint(self.getExtendedFingerprint(key))

    def contains(self, key):
        """
        key: ByteArrayType
        returns: bool
        """
        return self.containsExtendedFingerprint(self.getExtendedFingerprint(key))


    # -------------------------------------------------------------
    # Init/deinit like stuff.
    # -------------------------------------------------------------

    def clear(self):
        """
        returns: None
        """
        self.victim = CuckooFilter.ExtendedFingerprint(0,0,0)
        self.bucket_array = [EMPTY] * CuckooFilter.getfingerprintcount(self.bucketCount(), self.nests_per_bucket)

    # -------------------------------------------------------------
    # Size stuff.
    # -------------------------------------------------------------

    @staticmethod
    def sizeof(typ):
        size_dict = {
            'uint8': 1,
            'uint16': 2,
            'uint32': 4,
            'uint64': 8,
            'CuckooFilter': 1 + 1 + 2 + 1 + 1,
            'FingerprintType': 2,
            'IndexType': 1
        }
        if typ in size_dict:
            return size_dict[typ]
        return -1

    @staticmethod
    def getfingerprintcount(bucket_count, nests_per_bucket):
        """
        bucket_count: IndexType
        nests_per_bucket: IndexType
        returns: int
        """
        return bucket_count * nests_per_bucket

    @staticmethod
    def getbuffersize(bucket_count, nests_per_bucket):
        """
        bucket_count: IndexType
        nests_per_bucket: IndexType
        returns: int
        """
        return CuckooFilter.getfingerprintcount(bucket_count, nests_per_bucket) * CuckooFilter.sizeof('FingerprintType')

    @staticmethod
    def getsize(bucket_count, nests_per_bucket):
        """
        bucket_count: IndexType
        nests_per_bucket: IndexType
        returns: int
        """
        return CuckooFilter.sizeof('CuckooFilter') + CuckooFilter.getbuffersize(bucket_count, nests_per_bucket)

    def fingerprintcount(self):
        """
        returns: int
        """
        return CuckooFilter.getfingerprintcount(self.bucketCount(), self.nests_per_bucket)

    def bucketCount(self):
        return 1 << self.bucketCountLog2

    def buffersize(self):
        """
        returns: int
        """
        return CuckooFilter.getbuffersize(self.bucketCount(), self.nests_per_bucket)

    def size(self):
        """
        returns: int
        """
        return CuckooFilter.getsize(self.bucketCount(), self.nests_per_bucket)


    def saturate(self):
        self.bucket_array = self._cleanBucketArray(self.bucket_array[0])

    def _cleanBucketArray(self, replaceValue = 0):
        actual_bucket_array = []
        for item in self.bucket_array:
            if item == EMPTY:
                actual_bucket_array.append(replaceValue)
            else:
                actual_bucket_array.append(item)
        return actual_bucket_array



import time
from crownstone_core.util.Timestamp         import reconstructTimestamp
from crownstone_core.protocol.SwitchState   import SwitchState
from crownstone_core.util.BufferReader       import BufferReader
from crownstone_core.packets.serviceDataParsers.containers.AdvCrownstoneState import AdvCrownstoneState
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvFlags import AdvFlags
from crownstone_core.packets.serviceDataParsers.containers.AdvExternalCrownstoneState import AdvExternalCrownstoneState
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvGlobalFlags import AdvGlobalFlags


def parseStatePacket(reader: BufferReader):
    packet = AdvCrownstoneState()
    _parseStatePacket(packet, reader)
    packet.globalFlags = AdvGlobalFlags(reader.getUInt8())
    packet.validation = reader.getUInt8()
    return packet


def parseExternalStatePacket(reader: BufferReader):
    packet = AdvExternalCrownstoneState()
    _parseStatePacket(packet, reader)
    # position 13 on the state packet is the global flags, it is used in the external state for rssi
    packet.rssiOfExternalCrownstone = reader.getInt8()
    packet.validation = reader.getUInt8()
    return packet


def _parseStatePacket(packet, reader):
    packet.crownstoneId = reader.getUInt8()
    packet.switchState = SwitchState(reader.getUInt8())
    packet.flags = AdvFlags(reader.getUInt8())
    packet.temperature = reader.getInt8()
    powerFactor = reader.getInt8()
    realPower = reader.getInt16()

    packet.powerFactor = float(powerFactor) / 127.0

    # we cannot have a 0 for a power factor.To avoid division by 0, we set it to be either 0.01 or -0.01
    if 0 <= packet.powerFactor < 0.01:
        packet.powerFactor = 0.01
    elif -0.01 < packet.powerFactor < 0:
        packet.powerFactor = -0.01

    packet.powerUsageReal = float(realPower) / 8.0
    packet.powerUsageApparent = packet.powerUsageReal / packet.powerFactor

    packet.accumulatedEnergy = reader.getInt32()
    packet.accumulatedEnergy = packet.accumulatedEnergy * 64  # correction to display in J

    partialTimestamp        = reader.getUInt16()
    packet.uniqueIdentifier = partialTimestamp

    if packet.flags.timeIsSet:
        packet.timestamp = reconstructTimestamp(time.time(), partialTimestamp)
    else:
        packet.timestamp = partialTimestamp  # this is now a counter



import time

from crownstone_core.packets.serviceDataParsers.containers.elements.AdvCrownstoneErrorBitmask import AdvCrownstoneErrorBitmask
from crownstone_core.util.Timestamp           import reconstructTimestamp
from crownstone_core.util.BufferReader        import BufferReader
from crownstone_core.packets.serviceDataParsers.containers.AdvErrorPacket import AdvErrorPacket
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvFlags import AdvFlags
from crownstone_core.packets.serviceDataParsers.containers.AdvExternalErrorPacket import AdvExternalErrorPacket


def parseErrorPacket(reader: BufferReader):
    packet = AdvErrorPacket()
    _parseErrorPacket(packet, reader)
    realPower = reader.getInt16()
    packet.powerUsageReal = float(realPower) / 8.0
    return packet


def parseExternalErrorPacket(reader: BufferReader):
    packet = AdvExternalErrorPacket()
    _parseErrorPacket(packet, reader)
    packet.powerUsageReal = reader.getUInt8()
    packet.validation     = reader.getUInt8()
    return packet


def _parseErrorPacket(packet, reader: BufferReader):
    packet.crownstoneId     = reader.getUInt8()
    packet.errorsBitmask    = AdvCrownstoneErrorBitmask(reader.getUInt32())
    packet.errorTimestamp   = reader.getUInt32()
    packet.flags            = AdvFlags(reader.getUInt8())
    packet.temperature      = reader.getInt8()
    partialTimestamp        = reader.getUInt16()
    packet.uniqueIdentifier = partialTimestamp

    if packet.flags.timeIsSet:
        packet.timestamp = reconstructTimestamp(time.time(), partialTimestamp)
    else:
        packet.timestamp = partialTimestamp  # this is now a counter



import logging
import traceback
import uuid
from typing import Callable, Any

_LOGGER = logging.getLogger(__name__)

class EventBus:

    def __init__(self):
        self.topics = {}
        self.subscriberIds = {}


    def once(self, topic: str, callback: Callable[[Any], None]) -> str:
        def cleanup(bus, subId, data):
            if subscriptionId in self.subscriberIds:
                callback(data)
            bus.unsubscribe(subId)

        subscriptionId = self.subscribe(topic, lambda data: cleanup(self, subscriptionId, data))
        return subscriptionId

    def subscribe(self, topic: str, callback: Callable[[Any], None]) -> str:
        # Returns a subscriptionId to be used to unsubscribe.
        if topic not in self.topics:
            self.topics[topic] = {}

        subscriptionId = str(uuid.uuid4())
        self.subscriberIds[subscriptionId] = topic
        self.topics[topic][subscriptionId] = callback

        return subscriptionId

    def emit(self, topic: str, data: Any = None):
        if topic in self.topics:
            callbackIds = list(self.topics[topic].keys())
            for subscriptionId in callbackIds:
                try:
                    self.topics[topic][subscriptionId](data)
                except Exception as e:
                    # TODO: maybe kill the process, or raise a new exception?
                    _LOGGER.error(f"Error in callback of subscriptionId={subscriptionId}: {e}")
                    _LOGGER.info(traceback.format_exc())


    def unsubscribe(self, subscriptionId: str):
        if subscriptionId is None:
            return

        if subscriptionId in self.subscriberIds:
            topic = self.subscriberIds[subscriptionId]
            if topic in self.topics:
                self.topics[topic].pop(subscriptionId)

            self.subscriberIds.pop(subscriptionId)
        else:
            pass




from enum import IntEnum

from crownstone_core.util.BufferWriter import BufferWriter
from crownstone_core.packets.BasePacket import BasePacket


class InputDescriptionType(IntEnum):
	"""
	Describes a selection of data from an asset advertisement.
	"""
	MAC_ADDRESS = 0
	AD_DATA = 1
	MASKED_AD_DATA = 2


class InputDescriptionPacket(BasePacket):
	def __init__(self, type: InputDescriptionType):
		self.type = type

	def _serialize(self, writer: BufferWriter):
		writer.putUInt8(self.type)


class InputDescriptionMacAddress(InputDescriptionPacket):
	def __init__(self):
		super().__init__(InputDescriptionType.MAC_ADDRESS)

	def __str__(self):
		return f"InputDescriptionMacAddress(" \
		       f"type={self.type})"


class InputDescriptionFullAdData(InputDescriptionPacket):
	def __init__(self, adType: int):
		super().__init__(InputDescriptionType.AD_DATA)
		self.adType = adType

	def _serialize(self, writer: BufferWriter):
		super()._serialize(writer)
		writer.putUInt8(self.adType)

	def __str__(self):
		return f"InputDescriptionFullAdData(" \
		       f"type={self.type} " \
		       f"adType={self.adType})"


class InputDescriptionMaskedAdData(InputDescriptionPacket):
	def __init__(self, adType: int, mask: int):
		super().__init__(InputDescriptionType.MASKED_AD_DATA)
		self.adType = adType
		self.mask = mask

	def _serialize(self, writer: BufferWriter):
		super()._serialize(writer)
		writer.putUInt8(self.adType)
		writer.putUInt32(self.mask)

	def __str__(self):
		return f"InputDescriptionMaskedAdData(" \
		       f"type={self.type} " \
		       f"adType={self.adType} " \
		       f"mask={self.mask:032b})"


from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvUnknownData:
    def __init__(self):
        self.type = AdvType.UNKNOWN_DATA
        self.data       = None

    def __str__(self):
        return f"{self.type}\n"\
           f"data:       {self.data       }\n"



import time
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.Timestamp import reconstructTimestamp
from crownstone_core.packets.serviceDataParsers.containers.AdvMicroappData import AdvMicroappData
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvMicroappFlags import AdvMicroappFlags


def parseMicroappServiceData(reader: BufferReader):
    packet = AdvMicroappData()

    packet.flags = AdvMicroappFlags(reader.getUInt8())
    packet.microappUuid     = reader.getUInt16()
    packet.microappData     = reader.getBytes(8)
    packet.crownstoneId     = reader.getUInt8()
    packet.partialTimestamp = reader.getUInt16()
    packet.validation       = reader.getUInt8()

    packet.uniqueIdentifier = packet.partialTimestamp
    if packet.flags.timeIsSet:
        packet.timestamp = reconstructTimestamp(time.time(), packet.partialTimestamp)
    else:
        packet.timestamp = packet.partialTimestamp # this is now a counter

    return packet



from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvCrownstoneSetupState:
    def __init__(self):
        self.type = AdvType.SETUP_STATE

        self.switchState               = None
        self.flags                     = None
        self.temperature               = None
        self.powerFactor               = None
        self.powerUsageReal            = None
        self.powerUsageApparent        = None
        self.errorsBitmask             = None
        self.uniqueIdentifier          = None

    def __str__(self):
        return f"{self.type}\n"\
           f"switchState:        {self.switchState       }\n" \
           f"flags:            \n{self.flags             }\n" \
           f"temperature:        {self.temperature       }\n" \
           f"powerFactor:        {self.powerFactor       }\n" \
           f"powerUsageReal:     {self.powerUsageReal    }\n" \
           f"powerUsageApparent: {self.powerUsageApparent}\n" \
           f"errorBitmask:       {self.errorsBitmask     }\n" \
           f"uniqueIdentifier:   {self.uniqueIdentifier  }\n"


from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.microapp.MicroappSdkVersionPacket import MicroappSdkVersionPacket
from crownstone_core.packets.microapp.MicroappTestsPacket import MicroappTestsPacket
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter


class MicroappStatusPacket(BasePacket):
	def __init__(self, data=None):
		self.buildVersion = 0
		self.sdkVersion = MicroappSdkVersionPacket()
		self.checksum = 0
		self.checksumHeader = 0
		self.tests = MicroappTestsPacket()
		self.functionTrying = 0
		self.functionFailed = 0
		self.functionsPassed = []

		if data is not None:
			self.deserialize(data)

	def _deserialize(self, reader: BufferReader):
		self.buildVersion = reader.getUInt32()
		self.sdkVersion.deserialize(reader)
		self.checksum = reader.getUInt16()
		self.checksumHeader = reader.getUInt16()
		self.tests.deserialize(reader)
		self.functionTrying = reader.getUInt8()
		self.functionFailed = reader.getUInt8()

		self.functionsPassed = []
		functionsPassed = reader.getUInt32()
		for i in range(0, 32):
			if functionsPassed & (1 << i):
				self.functionsPassed.append(i)

	def __str__(self):
		return f"MicroappStatusPacket(" \
		       f"buildVersion={self.buildVersion}, " \
		       f"sdkVersion={self.sdkVersion}, " \
		       f"checksum={self.checksum}, " \
		       f"tests={self.tests}, " \
		       f"functionTrying={self.functionTrying}, " \
		       f"functionFailed={self.functionFailed}, " \
		       f"functionsPassed={self.functionsPassed})"



from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvExternalCrownstoneState:
    def __init__(self):
        self.type = AdvType.EXTERNAL_STATE

        self.crownstoneId              = None
        self.switchState               = None
        self.flags                     = None
        self.temperature               = None
        self.powerFactor               = None
        self.powerUsageReal            = None
        self.powerUsageApparent        = None
        self.accumulatedEnergy         = None
        self.rssiOfExternalCrownstone  = None
        self.timestamp                 = None
        self.uniqueIdentifier          = None
        self.validation                = None


    def __str__(self):
        return f"{self.type}\n"\
           f"crownstoneId:             {self.crownstoneId            }\n" \
           f"switchState:              {self.switchState             }\n" \
           f"flags:                  \n{self.flags                   }\n" \
           f"temperature:              {self.temperature             }\n" \
           f"powerFactor:              {self.powerFactor             }\n" \
           f"powerUsageReal:           {self.powerUsageReal          }\n" \
           f"powerUsageApparent:       {self.powerUsageApparent      }\n" \
           f"accumulatedEnergy:        {self.accumulatedEnergy       }\n" \
           f"rssiOfExternalCrownstone: {self.rssiOfExternalCrownstone}\n" \
           f"timestamp:                {self.timestamp               }\n" \
           f"uniqueIdentifier:         {self.uniqueIdentifier        }\n" \
           f"validation:               {self.validation              }\n"



from crownstone_core.util import BufferWriter

from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.assetFilter.FilterDataPackets import FilterData


class CuckooExtendedFingerprint(BasePacket):
	def __init__(self):
		self.fingerprint = 0
		self.bucketA = 0
		self.bucketB = 0

	def _serialize(self, writer: BufferWriter):
		writer.putUInt16(self.fingerprint)
		writer.putUInt8(self.bucketA)
		writer.putUInt8(self.bucketB)

	def __str__(self):
		return f"CuckooExtendedFingerprint(" \
		       f"fingerprint={self.fingerprint} " \
		       f"bucketA={self.bucketA} " \
		       f"bucketB={self.bucketB})"



class CuckooFilterData(FilterData):
	def __init__(self):
		self.bucketCountLog2: int = 0
		self.nestsPerBucket: int = 0
		self.victim = CuckooExtendedFingerprint()
		self.bucketArray = []

	def _serialize(self, writer: BufferWriter):
		writer.putUInt8(self.bucketCountLog2)
		writer.putUInt8(self.nestsPerBucket)
		self.victim.serialize(writer)
		for bucket in self.bucketArray:
			writer.putUInt16(bucket)

	def __str__(self):
		return f"CuckooFilterData(" \
		       f"bucketCountLog2={self.bucketCountLog2} " \
		       f"nestsPerBucket={self.nestsPerBucket} " \
		       f"victim={self.victim} " \
		       f"bucketArray={self.bucketArray})"



__author__ = 'Bart van Vliet'

import binascii
import struct

class Conversion:

    ####################################
    # Conversions between uint and int #
    ####################################
    @staticmethod
    def uint8_to_int8(byte):
        """	Convert a uint8 to an int8 """
        res = byte
        if res > 127:
            res -= 256
        return res

    @staticmethod
    def int8_to_uint8(byte):
        """	Convert an int8 to a uint8 """
        res = byte
        if res < 0:
            res += 256
        return res

    @staticmethod
    def uint16_to_int16(val):
        """ Convert a uint16 to an int16 """
        res = val
        if res > 32767:
            res -= 65536
        return res

    @staticmethod
    def int16_to_uint16(val):
        """ Convert an int16 to a uint16 """
        res = val
        if res < 0:
            res += 65536
        return res

    @staticmethod
    def uint32_to_int32(val):
        """ Convert an uint32 to a int32 """
        res = val
        if res > 2147483647:
            res -= 4294967296
        return res

    @staticmethod
    def int32_to_uint32(val):
        """ Convert an int32 to a uint32 """
        res = val
        if res < 0:
            res += 4294967296
        return res

    #######################################
    # Conversions between uint8 and float #
    #######################################
    @staticmethod
    def uint8_array_to_float(arr8):
        """ Convert an array of 4 bytes to a float """
        return struct.unpack('<f', bytearray(arr8))[0]

    @staticmethod
    def float_to_uint8_array(val):
        """ Convert a float to an array of 4 bytes """
        return bytearray(struct.pack('<f', val))

    ########################################
    # Conversions between uint8 and uint16 #
    ########################################
    @staticmethod
    def uint8_to_uint16(byte0, byte1):
        """ Convert 2 bytes to a uint16 """
        return (byte1 << 8) + byte0

    @staticmethod
    def uint8_array_to_uint16(arr8):
        """ Convert an array of 2 bytes to a uint16 """
        return (arr8[1] << 8) + arr8[0]

    @staticmethod
    def uint8_array_to_int16(arr8):
        """ Convert an array of 2 bytes to a uint16 """
        return Conversion.uint16_to_int16(Conversion.uint8_array_to_uint16(arr8))

    @staticmethod
    def uint16_to_uint8_array(value):
        """ Convert a number into an array of 2 bytes. """
        return [
            (value >> 0 & 0xFF),
            (value >> 8 & 0xFF)
        ]

    @staticmethod
    def int16_to_uint8_array(value):
        return Conversion.uint16_to_uint8_array(Conversion.int16_to_uint16(value))

    @staticmethod
    def uint8_to_uint8_array(value):
        """ Convert a number into an array of 1 bytes. """
        return [ value & 0xFF ]

    @staticmethod
    def uint8_array_to_uint16_array(arr8):
        """	Convert a uint8 array to a uint16 array	"""
        arr16 = []
        for i in range(1, len(arr8), 2):
#			arr16.append(Conversion.uint8_array_to_uint16(arr8[i-1 : i+1]))
            arr16.append(Conversion.uint8_to_uint16(arr8[i-1], arr8[i]))
        return arr16

    @staticmethod
    def uint16_array_to_uint8_array(arr16):
        """	Convert a uint16 array to a byte array """
        arr8 = []
        for val in arr16:
            arr8.extend(Conversion.uint16_to_uint8_array(val))
        return arr8


    ########################################
    # Conversions between uint8 and uint32 #
    ########################################
    @staticmethod
    def uint8_to_uint32(byte0, byte1, byte2, byte3):
        """ Convert 4 bytes to a uint32 """
        return (byte3 << 24) + (byte2 << 16) + (byte1 << 8) + byte0

    @staticmethod
    def uint8_array_to_uint32(arr8):
        """ Convert an array of 4 bytes to a uint32 """
        return (arr8[3] << 24) + (arr8[2] << 16) + (arr8[1] << 8) + arr8[0]

    @staticmethod
    def uint8_array_to_uint64(arr8):
        """ Convert an array of 4 bytes to a uint32 """
        return (arr8[7] << 56) + \
               (arr8[6] << 48) + \
               (arr8[5] << 40) + \
               (arr8[4] << 32) + \
               (arr8[3] << 24) + \
               (arr8[2] << 16) + \
               (arr8[1] << 8) + \
               arr8[0]


    @staticmethod
    def uint8_array_to_int32(arr8):
        """ Convert an array of 4 bytes to a uint32 """
        return Conversion.uint32_to_int32(Conversion.uint8_array_to_uint32(arr8))

    @staticmethod
    def uint8_array_to_uint32_array(arr8):
        """	Convert a uint8 array to a uint16 array	"""
        arr32 = []
        for i in range(3, len(arr8), 4):
#			arr32.append(Conversion.uint8_array_to_uint32(arr8[i-3 : i+1]))
            arr32.append(Conversion.uint8_to_uint32(arr8[i-3], arr8[i-2], arr8[i-1], arr8[i]))
        return arr32

    @staticmethod
    def uint32_to_uint8_array(value):
        """ Convert a number into an array of 4 bytes. """
        return [
            (value >> 0 & 0xFF),
            (value >> 8 & 0xFF),
            (value >> 16 & 0xFF),
            (value >> 24 & 0xFF)
        ]
    @staticmethod
    def int32_to_uint8_array(value):
        """ Convert a number into an array of 4 bytes. """
        return Conversion.uint32_to_uint8_array(Conversion.int32_to_uint32(value))

    @staticmethod
    def uint64_to_uint8_array(value):
        """ Convert a number into an array of 8 bytes. (little endian) """
        return [ value >> 8*i & 0xFF for i in range(8)]

    @staticmethod
    def uint32_array_to_uint8_array(arr32):
        """	Convert a uint32 array to a byte array """
        arr8 = []
        for val in arr32:
            arr8.extend(Conversion.uint32_to_uint8_array(val))
        return arr8


    @staticmethod
    def uint32_to_uint16_reversed_array(uint32):
        return [
            uint32 >> 16,
            (uint32 >> 0) % (0xFFFF + 1),
        ]

    @staticmethod
    def uint16_reversed_array_to_uint32(uint16Array):
        return uint16Array[1] + (uint16Array[0] * (0xFFFF + 1))


    ##############################
    # uint8 array to/from string #
    ##############################
    @staticmethod
    def uint8_array_to_string(arr8):
#		string = ""
#		for i in range(0, len(arr8)):
#			string += chr(arr8[i])
#		return string
#		return str(bytearray(arr8))
        return bytearray(arr8).decode()

    @staticmethod
    def string_to_uint8_array(string):
        arr8 = []
        for i in range(0, len(string)):
            arr8.append(ord(string[i]))
        return bytes(arr8)

    @staticmethod
    def uint8_array_to_uint8(arr8):
        return int(arr8[0] & 0xFF)

    @staticmethod
    def uint8_array_to_int8(arr8):
        return Conversion.uint8_to_int8(Conversion.uint8_array_to_uint8(arr8))



    ########################
    # Number to hex string #
    ########################
    @staticmethod
    def uint8_to_hex_string(val):
        if val > 255 or val < 0:
            raise Exception(f"Value must be of type uint8. Value = {val}")
        hex_str = "%02x" % val
        return hex_str

    @staticmethod
    def uint8_array_to_hex_string(arr):
#		hex_str = Conversion.uint8_to_hex_string(arr[0])
#		for i in range(1, len(arr)):
#			hex_str += Conversion.uint8_to_hex_string(arr[i])
#		return hex_str
        return binascii.b2a_hex(bytearray(arr)).decode('utf-8')

    @staticmethod
    def uint16_to_hex_string(val):
        arr8 = Conversion.uint16_to_uint8_array(val)
        return Conversion.uint8_array_to_hex_string(arr8)

    @staticmethod
    def uint32_to_hex_string(val):
        arr8 = Conversion.uint32_to_uint8_array(val)
        return Conversion.uint8_array_to_hex_string(arr8)


    #######################
    # Hex string to array #
    #######################

    @staticmethod
    def ibeaconUUIDString_to_uint8_array(inputStr):
        hexStr = inputStr.replace(":","").replace("-","")
        return Conversion.hex_string_to_uint8_array(hexStr)

    @staticmethod
    def ibeaconUUIDString_to_reversed_uint8_array(inputStr):
        arr = Conversion.ibeaconUUIDString_to_uint8_array(inputStr)
        arr.reverse()
        return arr

    @staticmethod
    def ascii_or_hex_string_to_16_byte_array(inputStr):
        if len(inputStr) == 16:
            return Conversion.string_to_uint8_array(inputStr)
        else :
            return Conversion.hex_string_to_uint8_array(inputStr)


    @staticmethod
    def hex_string_to_uint8_array(hexStr):
#		""" Convert a string which represents a hex byte buffer to an uint8 array """
#		""" Only converts buffer from start to start+size bytes """
#		buf = buffer_str.split(" ")
#		if (size == False):
#			size=len(buf)
#		arr = []
#		for i in range(start, start+size):
#			arr.append(int(buf[i], 16))
#		return arr
        """
        Converts a hexadecimal string to a byte array
        :param hexStr: hexadecimal string to be converted
        :type hexStr: str
        :rtype: bytearray
        """
        return bytearray(binascii.a2b_hex(hexStr.encode('utf-8')))

    @staticmethod
    def hex_string_to_uint16_array(hexStr):
#	def hex_string_to_uint16_array(buffer_str, start=0, size=False):
#		""" Convert a string which represents a hex byte buffer to an uint16 array """
#		""" Only converts buffer from start to start+size*2 bytes """
#		buf = buffer_str.split(" ")
#		if (size == False):
#			size=len(buf)/2
#		arr16 = []
#		for i in range(0, size):
#			arr16.append(Conversion.uint8_array_to_uint16([int(buf[start+2*i], 16), int(buf[start+2*i+1], 16)]))
#		return arr16
        """
        Converts a hexadecimal string to a uint16 array
        :param hexStr: hexadecimal string to be converted
        :type hexStr: str
        :rtype: list
        """
        arr8 = bytearray(binascii.a2b_hex(hexStr.encode('utf-8')))
        return Conversion.uint8_array_to_uint16(arr8)

    @staticmethod
    def hex_string_to_uint32_array(hexStr):
#	def hex_string_to_uint32_array(buffer_str, start=0, size=False):
#		""" Convert a string which represents a hex byte buffer to an uint32 array """
#		""" Only converts buffer from start to start+size*4 bytes """
#		buf = buffer_str.split(" ")
#		if (size == False):
#			size=len(buf)/4
#		arr32 = []
#		for i in range(0, size):
#			arr32.append(Conversion.uint8_array_to_uint32([int(buf[start+4*i], 16), int(buf[start+4*i+1], 16), int(buf[start+4*i+2], 16), int(buf[start+4*i+3], 16)]))
#		return arr32
        """
        Converts a hexadecimal string to a uint32 array
        :param hexStr: hexadecimal string to be converted
        :type hexStr: str
        :rtype: list
        """
        arr8 = bytearray(binascii.a2b_hex(hexStr.encode('utf-8')))
        return Conversion.uint8_array_to_uint32(arr8)

    @staticmethod
    def address_to_uint8_array(addressStr):
        """
        Converts a bluetooth address string to a uint8 array
        :param addressStr: address string to be converted, must be like: "01:23:45:67:89:AB"
        :type addressStr: str
        :rtype: list
        """
        hexStrArr = addressStr.split(":")
        if len(hexStrArr) != 6:
            return []

        arr8 = []
        for p in reversed(hexStrArr):
            arr8.append(Conversion.hex_string_to_uint8_array(p)[0])
        return bytearray(arr8)

    @staticmethod
    def uint8_array_to_address(arr8):
        """
        Converts a uint8 array to a bluetooth address string
        :param arr8: array of uint8, should be 6 long.
        :type arr8: bytearray
        :rtype: str
        """
        if (len(arr8) != 6):
            return ""
        hexStr = Conversion.uint8_to_hex_string(arr8[-1])
        for b in reversed(arr8[0:-1]):
            hexStr = hexStr + ":" + Conversion.uint8_to_hex_string(b)
        return hexStr


    @staticmethod
    def bit_array_to_010_array(bitArray):
        result = [0] * len(bitArray)

        for i in range(0,len(bitArray)):
            if bitArray[i]:
                result[i] = 1

        return result



    @staticmethod
    def uint8_to_bit_array(val):
        result = [False, False, False, False, False, False, False, False]

        one = 1

        result[0] = (val & (one << 0)) != 0
        result[1] = (val & (one << 1)) != 0
        result[2] = (val & (one << 2)) != 0
        result[3] = (val & (one << 3)) != 0
        result[4] = (val & (one << 4)) != 0
        result[5] = (val & (one << 5)) != 0
        result[6] = (val & (one << 6)) != 0
        result[7] = (val & (one << 7)) != 0

        return result

    @staticmethod
    def uint16_to_bit_array(val):
        result = [False] * 16
        one = 1

        for i in range(0, 16):
            result[i] = (val & (one << 15 - i)) != 0

        return result

    @staticmethod
    def uint32_to_bit_array(val):
        result = [False] * 32
        one = 1

        for i in range(0,32):
            result[i] = (val & (one << 31 - i)) != 0

        return result

    @staticmethod
    def uint32_to_bit_array_reversed(val):
        result = [False] * 32
        one = 1

        for i in range(0,32):
            result[i] = (val & (one << 31 - i)) != 0

        return list(reversed(result))



from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter


class MicroappUploadPacket(BasePacket):
	def __init__(self, header: MicroappHeaderPacket, offset: int, binaryChunk: bytearray):
		self.header = header
		self.offset = offset
		self.binaryChunk = binaryChunk

	def _deserialize(self, reader: BufferReader):
		self.header.deserialize(reader)
		self.offset = reader.getUInt16()
		self.binaryChunk = reader.getRemainingBytes()

	def _serialize(self, writer: BufferWriter):
		self.header.serialize(writer)
		writer.putUInt16(self.offset)
		writer.putBytes(self.binaryChunk)

	def __str__(self):
		return f"MicroappUploadPacket(" \
		       f"offset={self.offset}, " \
		       f"binaryChunk={[self.binaryChunk]})"



from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvMicroappData:
    def __init__(self):
        self.type = AdvType.MICROAPP_DATA

        self.crownstoneId     = None
        self.microappUuid     = None
        self.microappData     = None
        self.flags            = None
        self.uniqueIdentifier = None
        self.validation       = None

    def __str__(self):
        return f"{self.type}\n"\
           f"crownstoneId     {self.crownstoneId    }\n" \
           f"microappUuid     {self.microappUuid    }\n" \
           f"microappData     {self.microappData    }\n" \
           f"flags            {self.flags           }\n" \
           f"uniqueIdentifier {self.uniqueIdentifier}\n" \
           f"validation       {self.validation      }\n"



import time
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.Timestamp import reconstructTimestamp
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvHubFlags import AdvHubFlags
from crownstone_core.packets.serviceDataParsers.containers.AdvHubState import AdvHubState


def parseHubData(reader: BufferReader):
    packet = AdvHubState()

    packet.crownstoneId = reader.getUInt8()
    packet.flags        = AdvHubFlags(reader.getUInt8())
    packet.hubData      = reader.getBytes(9)
    partialTimestamp    = reader.getUInt16()
    reader.skip()
    packet.validation   = reader.getUInt8()

    packet.uniqueIdentifier = partialTimestamp
    if packet.flags.timeIsSet:
        packet.timestamp = reconstructTimestamp(time.time(), partialTimestamp)
    else:
        packet.timestamp = partialTimestamp # this is now a counter

    return packet


from crownstone_core.packets.serviceDataParsers.serviceDataParsers.opCode6_type0_setupState        import parseSetupState
from crownstone_core.packets.serviceDataParsers.serviceDataParsers.opCode7_type4_alternative_state import parseAlternativeState
from crownstone_core.packets.serviceDataParsers.serviceDataParsers.opCode7_type6_microapp          import parseMicroappServiceData
from crownstone_core.packets.serviceDataParsers.serviceDataParsers.opCode7_type0_state_packet      import parseStatePacket, parseExternalStatePacket
from crownstone_core.packets.serviceDataParsers.serviceDataParsers.opCode7_type1_error_packet      import parseErrorPacket, parseExternalErrorPacket
from crownstone_core.packets.serviceDataParsers.serviceDataParsers.opCode7_type5_hubData           import parseHubData
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.Exceptions import CrownstoneException, CrownstoneError


def parseOpCode6(data):
    reader = BufferReader(data)
    dataType = reader.getUInt8()
    
    if dataType == 0:
        return parseSetupState(reader)
    elif dataType == 5:
        return parseHubData(reader)
    else:
        raise CrownstoneException(CrownstoneError.UNKNOWN_SERVICE_DATA, "Could not parse this dataType")
        


def parseOpcode7(data):
    reader = BufferReader(data)
    dataType = reader.getUInt8()

    if dataType == 0:
        return parseStatePacket(reader)
    elif dataType == 1:
        return parseErrorPacket(reader)
    elif dataType == 2:
        return parseExternalStatePacket(reader)
    elif dataType == 3:
        return parseExternalErrorPacket(reader)
    elif dataType == 4:
        return parseAlternativeState(reader)
    elif dataType == 5:
        return parseHubData(reader)
    elif dataType == 6:
        return parseMicroappServiceData(reader)
    else:
        raise CrownstoneException(CrownstoneError.UNKNOWN_SERVICE_DATA, "Could not parse this dataType")




from enum import Enum


class AdvType(Enum):
    CROWNSTONE_STATE   = 0
    CROWNSTONE_ERROR   = 1
    EXTERNAL_STATE     = 2
    EXTERNAL_ERROR     = 3
    ALTERNATIVE_STATE  = 4
    HUB_STATE          = 5
    MICROAPP_DATA      = 6

    SETUP_STATE        = 100

    UNKNOWN_DATA       = 200

    CROWNSTONE_FLAGS   = 300
    HUB_FLAGS          = 301
    MICROAPP_FLAGS     = 302

    CROWNSTONE_ERRORS  = 500


import logging
from typing import List

from crownstone_core.packets.assetFilter.util import AssetFilterMasterCrc

from crownstone_core.packets.assetFilter.FilterCommandPackets import FilterSummariesPacket
from crownstone_core.packets.assetFilter.builders.AssetFilter import AssetFilter

_LOGGER = logging.getLogger(__name__)

class AssetFilterSyncer:
    def __init__(self, summaries: FilterSummariesPacket, filters: List[AssetFilter], masterVersion: int = None):
        """
        Helper class to upload given filters, and remove other filters, and determines if a commit is required at all.

        After construction, read out the class variables.

        :param summaries:         The filter summaries packet of the Crownstone.
        :param filters:           The filters to be set at the Crownstone.
        :param masterVersion:     The new master version. If None, the master version will be increased by 1.
        """
        self.commitRequired = False
        self.removeIds = []
        self.uploadIds = []
        self.masterVersion = 0

        # Checks for duplicate IDs.
        masterCrc = AssetFilterMasterCrc.get_master_crc_from_filters(filters)

        # Iterate over the filters on the Crownstone, to see if there are abundant IDs, or filter CRC mismatches.
        for summary in summaries.summaries:
            foundId = False
            for filter in filters:
                if filter.getFilterId() == summary.id:
                    foundId = True
                    if filter.getCrc() != summary.crc:
                        _LOGGER.info(f"CRC mismatch, upload filderId={summary.id}")
                        self.uploadIds.append(summary.id)
                    else:
                        _LOGGER.info(f"CRC match, skip filderId={summary.id}")
            if not foundId:
                _LOGGER.info(f"Not found, remove filderId={summary.id}")
                self.removeIds.append(summary.id)

        # Iterate over given filters, to see if there are IDs that are not on the Crownstone.
        for filter in filters:
            foundId = False
            for summary in summaries.summaries:
                if filter.getFilterId() == summary.id:
                    foundId = True
            if not foundId:
                _LOGGER.info(f"Not found, upload filderId={filter.getFilterId()}")
                self.uploadIds.append(filter.getFilterId())

        # Only increase master version if there are any changes.
        if len(self.removeIds) == 0 \
                and len(self.uploadIds) == 0 \
                and masterCrc == summaries.masterCrc \
                and (masterVersion is None or summaries.masterVersion == masterVersion):
            _LOGGER.info("No changes, no commit requried.")
            self.commitRequired = False
            self.masterVersion = summaries.masterVersion
            return

        if masterVersion is None:
            self.masterVersion = summaries.masterVersion + 1
        else:
            self.masterVersion = masterVersion
        self.commitRequired = True


from crownstone_core.util.Conversion import Conversion
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType

class AdvHubFlags:

    def __init__(self, byte):
        self.type = AdvType.ALTERNATIVE_STATE

        bitmaskArray = Conversion.uint8_to_bit_array(byte)

        self.uartAlive                          = bitmaskArray[0]
        self.uartAliveEncrypted                 = bitmaskArray[1]
        self.uartEncryptionRequiredByCrownstone = bitmaskArray[2]
        self.uartEncryptionRequiredByHub        = bitmaskArray[3]
        self.hubHasBeenSetUp                    = bitmaskArray[4]
        self.hubHasInternet                     = bitmaskArray[5]
        self.hubHasError                        = bitmaskArray[6]
        self.timeIsSet                          = bitmaskArray[7]

    def __str__(self):
        return \
           f"    uartAlive:                          {self.uartAlive                         }\n" \
           f"    uartAliveEncrypted:                 {self.uartAliveEncrypted                }\n" \
           f"    uartEncryptionRequiredByCrownstone: {self.uartEncryptionRequiredByCrownstone}\n" \
           f"    uartEncryptionRequiredByHub:        {self.uartEncryptionRequiredByHub       }\n" \
           f"    hubHasBeenSetUp:                    {self.hubHasBeenSetUp                   }\n" \
           f"    hubHasInternet:                     {self.hubHasInternet                    }\n" \
           f"    hubHasError:                        {self.hubHasError                       }\n" \
           f"    timeIsSet:                          {self.timeIsSet                         }"


from crownstone_core.Enums import CrownstoneOperationMode

from crownstone_core.packets.ServiceData import ServiceData
from crownstone_core.protocol.Services import DFU_ADVERTISEMENT_SERVICE_UUID
from crownstone_core.util.Conversion import Conversion


class Advertisement:
    
    def __init__(self, address, rssi, nameText, serviceData, serviceUUID=None):
        self.address = address
        self.rssi = rssi
        self.name = nameText

        usedUUID = serviceUUID
        if isinstance(serviceUUID, str):
            usedUUID = serviceUUID.lower()

        self.serviceUUID = usedUUID
        self.serviceData = None
        self.operationMode = CrownstoneOperationMode.UNKNOWN

        dataString = serviceData
        
        if serviceData is not None:
            if isinstance(serviceData, str):
                dataArray = Conversion.hex_string_to_uint8_array(dataString)
                self.serviceUUID = Conversion.uint8_array_to_uint16([dataArray[0], dataArray[1]])
                # pop the service UUID
                dataArray.pop(0)
                dataArray.pop(0)
            else:
                dataArray = serviceData

            if dataArray:
                self.serviceData = ServiceData(dataArray)

            if self.serviceUUID == DFU_ADVERTISEMENT_SERVICE_UUID:
                self.operationMode = CrownstoneOperationMode.DFU
            else:
                self.operationMode = self.serviceData.getOperationMode()


    def isCrownstoneFamily(self):
        return self.serviceUUID == 0xC001 or self.serviceUUID == DFU_ADVERTISEMENT_SERVICE_UUID

    def hasServiceData(self):
        return self.serviceData is not None
    
    def getCrownstoneId(self):
        if self.hasServiceData() and self.isCrownstoneFamily():
            return self.serviceData.payload.crownstoneId
    
    def parse(self, decryptionKey = None):
        if self.hasServiceData() and self.isCrownstoneFamily():
            return self.serviceData.parse(decryptionKey = decryptionKey)

    # def getDictionary(self):
    #     data = {}
    #
    #     data["name"] = self.name
    #     data["rssi"] = self.rssi
    #     data["address"] = self.address
    #
    #     if self.serviceUUID is not None:
    #         data["serviceUUID"] = self.serviceUUID
    #
    #     if self.serviceData is not None:
    #         data["serviceData"] = self.serviceData.getDictionary()
    #
    #     return data
    #
    # def getSummary(self):
    #     data = {}
    #     if self.serviceData is not None:
    #         data = self.serviceData.getSummary(self.address)
    #     return data

    



import time

from crownstone_core.util.Conversion import Conversion


def obtainTimestamp(fullTimeStamp, lsbTimestamp):
	"""
	Combines the full timestamp with a partial timestamp.

	:param fullTimeStamp:         Full unix timestamp.
	:param lsbTimestamp:          Partial (least significant) timestamp, as a uint16.
	:returns:                     Unix timestamp with the least significant bytes set to the partial timestamp.
	"""
	restoredTimestamp = (int(fullTimeStamp) & 0xFFFF0000) + (int(lsbTimestamp) & 0x0000FFFF)
	return restoredTimestamp


def reconstructTimestamp(currentTimestamp, lsbTimestamp):
	"""
	Reconstructs a timestamp from a partial timestamp, that only has the least significant bytes.

	:param currentTimestamp:      Current time, obtained with time.time().
	:param lsbTimestamp:          Partial (least significant) timestamp, as a uint16.
	:returns:                     Reconstructed unix timestamp.
	"""
	# embed location data in the timestamp
	correctedTimestamp = getCorrectedLocalTimestamp(currentTimestamp)

	# attempt restoration
	restoredTimestamp = obtainTimestamp(correctedTimestamp, lsbTimestamp)

	halfUInt16 = 0x7FFF # roughly 9 hours in seconds

	# correct for overflows, check for drift from current time
	delta = correctedTimestamp - restoredTimestamp

	if -halfUInt16 < delta < halfUInt16:
		return restoredTimestamp
	elif delta < -halfUInt16:
		restoredTimestamp = obtainTimestamp(correctedTimestamp - 0xFFFF, lsbTimestamp)
	elif delta > halfUInt16:
		restoredTimestamp = obtainTimestamp(correctedTimestamp + 0xFFFF, lsbTimestamp)

	return restoredTimestamp

def getCorrectedLocalTimestamp(currentTimestamp):
	"""
	Get a timestamp in seconds from epoch, set to the local time.

	:param currentTimestamp:      Current time, obtained with time.time().
	:returns:                     Local unix timestamp.
	"""
	# This does not work, as mktime() ignores the daylight saving time:
	#     secondsFromGMT = round(time.time() - time.mktime(time.gmtime()))
	secondsFromGMT = time.localtime().tm_gmtoff
	correctedTimestamp = currentTimestamp + secondsFromGMT
	return correctedTimestamp


from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvAlternativeState:
    def __init__(self):
        self.type = AdvType.ALTERNATIVE_STATE

        self.crownstoneId             = None
        self.switchState              = None
        self.flags                    = None
        self.behaviourMasterHash      = None
        self.assetFilterMasterVersion = None
        self.assetFilterMasterCRC     = None
        self.timestamp                = None
        self.uniqueIdentifier         = None
        self.validation               = None

    def __str__(self):
        return f"{self.type}\n"\
           f"crownstoneId              {self.crownstoneId       }\n" \
           f"switchState               {self.switchState        }\n" \
           f"flags                     \n{self.flags            }\n" \
           f"behaviourMasterHash       {self.behaviourMasterHash}\n" \
           f"assetFilterMasterVersion  {self.assetFilterMasterVersion}\n" \
           f"assetFilterMasterCRC      {self.assetFilterMasterCRC}\n" \
           f"timestamp                 {self.timestamp          }\n" \
           f"uniqueIdentifier          {self.uniqueIdentifier   }\n" \
           f"validation                {self.validation         }\n"


from crownstone_core.Enums import CrownstoneOperationMode

from crownstone_core.Exceptions import CrownstoneException, CrownstoneError

from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.packets.serviceDataParsers.parsers import parseOpCode6, parseOpcode7
from crownstone_core.protocol.BluenetTypes import DeviceType
from crownstone_core.util.EncryptionHandler import EncryptionHandler
from crownstone_core.packets.serviceDataParsers.containers.AdvUnknownData import AdvUnknownData


class ServiceData:
    
    def __init__(self, data):
        """
        The parsed data ends up in the self.payload. These will be one of the Adv* classes defined in the ./containers/ folder.
        """
        self.opCode        = 0
        self.deviceType    = DeviceType.UNDEFINED
        self.operationMode = CrownstoneOperationMode.UNKNOWN
        self.payload       = None
        self.decrypted     = False

        self._data = data

        reader = BufferReader(self._data)
        self.opCode = reader.getUInt8()
        if self.opCode == 7:
            self.operationMode = CrownstoneOperationMode.NORMAL
        elif self.opCode == 6:
            self.operationMode = CrownstoneOperationMode.SETUP

        deviceType = reader.getUInt8()
        if DeviceType.has_value(deviceType):
            self.deviceType = DeviceType(deviceType)

    def parse(self, decryptionKey = None):
        if decryptionKey is not None:
            self.decrypt(decryptionKey)

        reader = BufferReader(self._data)
        reader.skip(2)

        if self.opCode == 7:
            try:
                self.payload = parseOpcode7(reader.getRemainingBytes())
            except CrownstoneException:
                self.payload = AdvUnknownData()
                self.payload.data = self._data
                raise CrownstoneException(CrownstoneError.INVALID_SERVICE_DATA, "Protocol not supported. Unknown data type. Could be because decryption failed.")
        elif self.opCode == 6:
            self.payload = parseOpCode6(reader.getRemainingBytes())
        else:
            self.payload = AdvUnknownData()
            self.payload.data = self._data
            raise CrownstoneException(CrownstoneError.INVALID_SERVICE_DATA, "Protocol not supported. Unknown opcode.")


    def getOperationMode(self):
        return self.operationMode

    def decrypt(self, keyHexString):
        if len(keyHexString) >= 16 and len(self._data) == 18:
            if self.operationMode == CrownstoneOperationMode.NORMAL:
                reader = BufferReader(self._data)
                encryptedData = reader.skip(2).getRemainingBytes()
                result = EncryptionHandler.decryptECB(encryptedData, keyHexString)

                for i in range(0, len(encryptedData)):
                    # the first 2 bytes are opcode and device type
                    self._data[i + 2] = result[i]

                self.decrypted = True
        else:
            raise CrownstoneException(CrownstoneError.COULD_NOT_DECRYPT, "ServiceData decryption failed. Invalid key or invalid data.")




"""
As we need consistent cross-platform pseudo random number generation. This is a port of the
Middle Square Weyl Sequence Random Number Generator found on https://mswsrng.wixsite.com/rand .
"""
from crownstone_core.util.Bitmasks import Bitmasks

class Msws:
    class State:
        def __init__(self, x , w, s):
            """
            x,w,s: int

            Note: s must be uneven for a proper Msws sequence
            """
            self.x = int(x) & Bitmasks.ff64
            self.w = int(w) & Bitmasks.ff64
            self.s = int(s) & Bitmasks.ff64

    def __init__(self, seed = None):
        """
        seed: int
        """
        if seed == None:
            # default seed
            self.state = Msws.State(0, 0, 0xb5ad4eceda1ce2a9)
        else:
            self.state = Msws._seed(seed)

    def get(self):
        """ Returns a random value and updates the self.state object. """
        return Msws._msws(self.state)

    @staticmethod
    def _msws(state):
        """
        Returns random value and updates the state objects internals values.

        state: State
        returns: int
        """
        state.x = (state.x * state.x) & Bitmasks.ff64
        state.w = (state.w + state.s) & Bitmasks.ff64
        state.x = (state.x + state.w) & Bitmasks.ff64
        state.x = ((state.x >> 32) | (state.x << 32)) & Bitmasks.ff64
        return state.x & Bitmasks.ff32

    seed_constants = [
        0x37e1c9b5e1a2b843, 0x56e9d7a3d6234c87, 0xc361be549a24e8c7, 0xd25b9768a1582d7b,
        0x18b2547d3de29b67, 0xc1752836875c29ad, 0x4e85ba61e814cd25, 0x17489dc6729386c1,
        0x7c1563ad89c2a65d, 0xcdb798e4ed82c675, 0xd98b72e4b4e682c1, 0xdacb7524e4b3927d,
        0x53a8e9d7d1b5c827, 0xe28459db142e98a7, 0x72c1b3461e4569db, 0x1864e2d745e3b169,
        0x6a2c143bdec97213, 0xb5e1d923d741a985, 0xb4875e967bc63d19, 0x92b64d5a82db4697,
        0x7cae812d896eb1a5, 0xb53827d41769542d, 0x6d89b42c68a31db5, 0x75e26d434e2986d5,
        0x7c82643d293cb865, 0x64c3bd82e8637a95, 0x2895c34d9dc83e61, 0xa7d58c34dea35721,
        0x3dbc5e687c8e61d5, 0xb468a235e6d2b193,
    ]

    @staticmethod
    def _seed(seed):
        """
        Returns an initial State object based on the given seed.

        seed: int
        returns: State
        """
        n = int(seed) & Bitmasks.ff64
        r = int(n / 100000000)
        t = int(n % 100000000)

        q = len(Msws.seed_constants)
        si = Msws.seed_constants[r % q] # must be uneven (true for given values in seed_constants)
        r//= q
        wi = t * si + r * si * 100000000
        xi = wi

        mswsi = Msws()
        mswsi.state = Msws.State(xi, wi, si)

        # get odd random number for low order digit
        u = (mswsi.get() % 8) * 2 + 1
        v = 1 << u

        # get rest of digits
        m = 60
        c = 0
        while m > 0:
            j = mswsi.get() # get 8 digit 32-bit random word

            for i in range(0,32,4):
                k = (j >> i) & 0xf # get a digit

                if k != 0 and (c & (1 << k)) == 0: # not 0 and not previous
                    c |= (1 << k)
                    u |= (k << m) # add digit to output
                    m -= 4
                    if m == 24 or m == 28:
                        c = (1 << k) | v
                    if m == 0:
                        break

        return Msws.State(u,u,u)

class RandomGenerator:
    def __init__(self, seed = None):
        """
        seed: int
        """
        self.engine = Msws(seed)

    def rand(self):
        """
        Return the next pseudo random number and update internal state.
        """
        return self.engine.get()

    def __call__(self):
        """
        Shorthand for self.rand()
        """
        return self.rand()




from typing import List

from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter
from crownstone_core.packets.BasePacket import BasePacket


ASSET_FILTER_PROTOCOL = 0

class UploadFilterChunkPacket(BasePacket):
    """
    Packet to upload a chunk of a filter.

    You can use the AssetFilterChunker to get chunks.
    """
    def __init__(self, filterId: int, chunkStartIndex: int, totalSize: int, chunkSize: int, chunk: list):
        self.commandProtocolVersion = ASSET_FILTER_PROTOCOL
        self.filterId = filterId
        self.chunkStartIndex = chunkStartIndex
        self.totalSize = totalSize
        self.chunkSize = chunkSize
        self.chunk = chunk

    def _serialize(self, writer: BufferWriter):
        writer.putUInt8(self.commandProtocolVersion)
        writer.putUInt8(self.filterId)
        writer.putUInt16(self.chunkStartIndex)
        writer.putUInt16(self.totalSize)
        writer.putUInt16(self.chunkSize)
        writer.putBytes(self.chunk)


class RemoveFilterPacket(BasePacket):
    def __init__(self, filterId: int):
        self.commandProtocolVersion = ASSET_FILTER_PROTOCOL
        self.filterId = filterId

    def _serialize(self, writer: BufferWriter):
        writer.putUInt8(self.commandProtocolVersion)
        writer.putUInt8(self.filterId)

class CommitFilterChangesPacket(BasePacket):
    def __init__(self, masterVersion: int, masterCrc: int):
        self.commandProtocolVersion = ASSET_FILTER_PROTOCOL
        self.masterVersion = masterVersion
        self.masterCrc = masterCrc

    def _serialize(self, writer: BufferWriter):
        writer.putUInt8(self.commandProtocolVersion)
        writer.putUInt16(self.masterVersion)
        writer.putUInt32(self.masterCrc)

class FilterSummariesPacket(BasePacket):
    def __init__(self, data=None):
        self.commandProtocolVersion = 0
        self.masterVersion = 0
        self.masterCrc = 0
        self.freeSpace = 0
        self.summaries: List[FilterSummaryPacket] = []
        if data is not None:
            self.deserialize(data)

    def _deserialize(self, reader: BufferReader):
        self.commandProtocolVersion = reader.getUInt8()
        self.masterVersion = reader.getUInt16()
        self.masterCrc = reader.getUInt32()
        self.freeSpace = reader.getUInt16()
        self.summaries.clear()
        while reader.getRemainingByteCount():
            self.summaries.append(FilterSummaryPacket(reader))

    def __str__(self):
        summariesString = "["
        for summary in self.summaries:
            summariesString += f"{summary}, "
        summariesString = summariesString[:-2] + "]"

        return f"FilterSummariesPacket(" \
               f"commandProtocolVersion={self.commandProtocolVersion} " \
               f"masterVersion={self.masterVersion} " \
               f"masterCrc={self.masterCrc} " \
               f"freeSpace={self.freeSpace} " \
               f"summaries={summariesString})"

class FilterSummaryPacket(BasePacket):
    def __init__(self, data=None):
        self.id = 0
        self.crc = 0
        if data is not None:
            self.deserialize(data)

    def _deserialize(self, reader: BufferReader):
        self.id = reader.getUInt8()
        self.crc = reader.getUInt32()

    def __str__(self):
        return f"FilterSummaryPacket(" \
               f"id={self.id} " \
               f"crc={self.crc})"



from enum import IntEnum

from crownstone_core.util.BufferWriter import BufferWriter

from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.assetFilter.InputDescriptionPackets import *

class FilterOutputDescriptionType(IntEnum):
    """
    What type of message should be output.
    """
    MAC_ADDRESS = 0
    ASSET_ID = 1
    NONE = 2
    ASSET_ID_NEAREST_CROWNSTONE = 100 # Experimental, may be removed or changed in a later release.

class FilterOutputDescription(BasePacket):
    def __init__(self,
                 outFormat: FilterOutputDescriptionType,
                 inFormat: None or InputDescriptionMacAddress or InputDescriptionFullAdData or InputDescriptionMaskedAdData):
        self.outFormat = outFormat
        self.inFormat  = inFormat

    def _serialize(self, writer: BufferWriter):
        writer.putUInt8(self.outFormat)
        if self.inFormat is not None:
            self.inFormat.serialize(writer)

    def __str__(self):
        return f"FilterOutputDescription(" \
               f"inFormat={self.inFormat} " \
               f"outFormat={self.outFormat})"



from crownstone_core.util.Conversion import Conversion
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType

class AdvGlobalFlags:

    def __init__(self, byte):
        self.type = AdvType.MICROAPP_FLAGS

        bitmaskArray = Conversion.uint8_to_bit_array(byte)

        self.behaviourEnabled = bitmaskArray[0]

    def __str__(self):
        return \
           f"    behaviourEnabled: {self.behaviourEnabled}"


from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvExternalErrorPacket:
    def __init__(self):
        self.type = AdvType.EXTERNAL_ERROR

        self.crownstoneId              = None
        self.errorsBitmask             = None
        self.errorTimestamp            = None
        self.flags                     = None
        self.temperature               = None
        self.rssiOfExternalCrownstone  = None
        self.timestamp                 = None
        self.uniqueIdentifier          = None
        self.validation                = None

    def __str__(self):
        return f"{self.type}\n"\
           f"crownstoneId             {self.crownstoneId            }\n" \
           f"errorsBitmask            {self.errorsBitmask           }\n" \
           f"errorTimestamp           {self.errorTimestamp          }\n" \
           f"flags                    {self.flags                   }\n" \
           f"temperature              {self.temperature             }\n" \
           f"rssiOfExternalCrownstone {self.rssiOfExternalCrownstone}\n" \
           f"timestamp                {self.timestamp               }\n" \
           f"uniqueIdentifier         {self.uniqueIdentifier        }\n" \
           f"validation               {self.validation              }\n"


import time

from crownstone_core.protocol.SwitchState import SwitchState
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.Timestamp import reconstructTimestamp
from crownstone_core.packets.serviceDataParsers.containers.AdvAlternativeState import AdvAlternativeState
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvFlags import AdvFlags


def parseAlternativeState(reader: BufferReader):
    packet = AdvAlternativeState()
    packet.crownstoneId             = reader.getUInt8()
    packet.switchState              = SwitchState(reader.getUInt8())
    packet.flags                    = AdvFlags(reader.getUInt8())
    packet.behaviourMasterHash      = reader.getUInt16()
    packet.assetFilterMasterVersion = reader.getUInt16()
    packet.assetFilterMasterCRC     = reader.getUInt32()
    partialTimestamp                = reader.getUInt16()
    reader.skip()
    packet.validation               = reader.getUInt8()
    packet.uniqueIdentifier         = partialTimestamp
    
    if packet.flags.timeIsSet:
        packet.timestamp = reconstructTimestamp(time.time(), partialTimestamp)
    else:
        packet.timestamp = partialTimestamp # this is now a counter

    return packet


def set_bit(bitmask: int, bit: int, flag: bool = True) -> int:
    """
    Set a bit in a bitmask.
    :param bitmask:     The initial bitmask.
    :param bit:         The bit to be set.
    :param flag:        True to set the bit high, False to set the bit low.
    :returns:           The updated bitmask.
    """
    if flag:
        bitmask |= (1 << bit)
    else:
        bitmask &= ~(1 << bit)
    return bitmask

def is_bit_set(bitmask: int, bit: int) -> bool:
    """
    Check if Nth bit is set in a bitmask.
    :param bitmask:     The bitmask.
    :param bit:         Which bit to check.
    :returns:           True when bit is set in bitmask.
    """
    return (bitmask & (1 << bit)) != 0

def get_bitmask(bits: list) -> int:
    """
    Returns a bitmask with given bits set.
    :param bits:        List of bits to be set.
    :returns:           The bitmask.
    """
    bitmask = 0
    for bit in bits:
        bitmask = set_bit(bitmask, bit)
    return bitmask

class Bitmasks:
    ff64 = 0xffffffffffffffff
    ff32 = 0xffffffff
    ff16 = 0xffff
    ff8  = 0xff


from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType
from crownstone_core.util.Conversion import Conversion


class AdvCrownstoneErrorBitmask:
    
    def __init__(self, bitMask):
        self.type = AdvType.CROWNSTONE_ERRORS

        self.bitMask = bitMask
    
        bitArray = Conversion.uint32_to_bit_array_reversed(bitMask)
    
        self.overCurrent        = bitArray[31 - 0]
        self.overCurrentDimmer  = bitArray[31 - 1]
        self.temperatureChip    = bitArray[31 - 2]
        self.temperatureDimmer  = bitArray[31 - 3]
        self.dimmerOnFailure    = bitArray[31 - 4]
        self.dimmerOffFailure   = bitArray[31 - 5]

    def __str__(self):
        return \
               f"    overCurrent:       {self.overCurrent       }\n"\
               f"    overCurrentDimmer: {self.overCurrentDimmer }\n"\
               f"    temperatureChip:   {self.temperatureChip   }\n"\
               f"    temperatureDimmer: {self.temperatureDimmer }\n"\
               f"    dimmerOnFailure:   {self.dimmerOnFailure   }\n"\
               f"    dimmerOffFailure:  {self.dimmerOffFailure  }\n"


from crownstone_core.util.Conversion import Conversion
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvFlags:

    def __init__(self, byte):
        self.type = AdvType.CROWNSTONE_FLAGS

        bitmaskArray = Conversion.uint8_to_bit_array(byte)

        self.dimmerReady         = bitmaskArray[0]
        self.dimmingAllowed      = bitmaskArray[1]
        self.hasError            = bitmaskArray[2]
        self.switchLocked        = bitmaskArray[3]
        self.timeIsSet           = bitmaskArray[4]
        self.switchCraftEnabled  = bitmaskArray[5]
        self.tapToToggleEnabled  = bitmaskArray[6]
        self.behaviourOverridden = bitmaskArray[7]

    def __str__(self):
        return \
           f"    dimmerReady:         {self.dimmerReady        }\n" \
           f"    dimmingAllowed:      {self.dimmingAllowed     }\n" \
           f"    hasError:            {self.hasError           }\n" \
           f"    switchLocked:        {self.switchLocked       }\n" \
           f"    timeIsSet:           {self.timeIsSet          }\n" \
           f"    switchCraftEnabled:  {self.switchCraftEnabled }\n" \
           f"    tapToToggleEnabled:  {self.tapToToggleEnabled }\n" \
           f"    behaviourOverridden: {self.behaviourOverridden}"



from crownstone_core.packets.assetFilter.FilterMetaDataPackets import FilterMetaData

from crownstone_core.packets.assetFilter.CuckooFilterPackets import CuckooFilterData
from crownstone_core.packets.assetFilter.ExactMatchFilter import ExactMatchFilter
from crownstone_core.packets.assetFilter.InputDescriptionPackets import *
from crownstone_core.util.BufferWriter import BufferWriter

class AssetFilterPacket(BasePacket):
    """
    A complete asset filter packet.
    It is recommended to use the AssetFilterBuilder to construct this class.
    """
    def __init__(self,
                 metaData: FilterMetaData = None,
                 filterData: CuckooFilterData or ExactMatchFilter = None):
        self.metaData = metaData
        self.filterData = filterData

    def _serialize(self, writer: BufferWriter):
        self.metaData.serialize(writer)
        self.filterData.serialize(writer)

    def __str__(self):
        return f"AssetFilter(" \
               f"metaData={self.metaData} " \
               f"filterData={self.filterData})"



import logging
from typing import List

from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.microapp.MicroappSdkVersionPacket import MicroappSdkVersionPacket
from crownstone_core.packets.microapp.MicroappStatusPacket import MicroappStatusPacket
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter

_LOGGER = logging.getLogger(__name__)

class MicroappInfoPacket(BasePacket):
	def __init__(self, data=None):
		self.protocol = 0
		self.maxApps = 0
		self.maxAppSize = 0
		self.maxChunkSize = 0
		self.maxRamUsage = 0
		self.sdkVersion = MicroappSdkVersionPacket()
		self.appsStatus: List[MicroappStatusPacket] = []

		if data is not None:
			self.deserialize(data)

	def _deserialize(self, reader: BufferReader):
		self.protocol = reader.getUInt8()
		self.maxApps = reader.getUInt8()
		self.maxAppSize = reader.getUInt16()
		self.maxChunkSize = reader.getUInt16()
		self.maxRamUsage = reader.getUInt16()
		self.sdkVersion.deserialize(reader)

		self.appsStatus = []
		for i in range(0, self.maxApps):
			statusPacket = MicroappStatusPacket()
			statusPacket.deserialize(reader)
			self.appsStatus.append(statusPacket)

	def __str__(self):
		appsStatusString = "["
		for status in self.appsStatus:
			appsStatusString += f"{status}, "
		appsStatusString = appsStatusString[:-2] + "]"

		return f"MicroappInfoPacket(" \
		       f"protocol={self.protocol}, " \
		       f"maxApps={self.maxApps}, " \
		       f"maxAppSize={self.maxAppSize}, " \
		       f"maxChunkSize={self.maxChunkSize}, " \
		       f"maxRamUsage={self.maxRamUsage}, " \
		       f"sdkVersion={self.sdkVersion}, " \
		       f"appsStatus={appsStatusString})"



from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter


class MicroappSdkVersionPacket(BasePacket):
	def __init__(self, data=None):
		self.major = 0
		self.minor = 0

		if data is not None:
			self.deserialize(data)

	def _deserialize(self, reader: BufferReader):
		self.major = reader.getUInt8()
		self.minor = reader.getUInt8()

	def _serialize(self, writer: BufferWriter):
		writer.putUInt8(self.major)
		writer.putUInt8(self.minor)

	def __str__(self):
		return f"{self.major}.{self.minor}"



from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter


class BasePacket:
	def _deserialize(self, reader: BufferReader):
		"""
		The deserialization function to be implemented by derived classes.
		:param reader: The class with data to be parsed.
		"""
		raise NotImplementedError

	def _serialize(self, writer: BufferWriter):
		"""
		The serialization function to be implemented by derived classes.
		:param writer: The class to write the data to.
		"""
		raise NotImplementedError

	def deserialize(self, data):
		if isinstance(data, BufferReader):
			self._deserialize(data)
		elif isinstance(data, list):
			reader = BufferReader(data)
			self._deserialize(reader)
		else:
			raise TypeError(f"Invalid type: {type(data)}")

	def serialize(self, data = None) -> list:
		if isinstance(data, BufferWriter):
			writer = data
			self._serialize(writer)
			buf = writer.getBuffer()
			return buf
		else:
			writer = BufferWriter()
			self._serialize(writer)
			buf = writer.getBuffer()
			if data is None:
				return buf
			else:
				data.extend(buf)
				return data



import logging
import math
from enum import Enum
from typing import List

from crownstone_core import Conversion
from crownstone_core.Exceptions import CrownstoneException, CrownstoneError
from crownstone_core.util.Bitmasks import set_bit, get_bitmask
from crownstone_core.util.CRC import crc32

from crownstone_core.util.Cuckoofilter import CuckooFilter

from crownstone_core.packets.assetFilter.AssetFilterPackets import AssetFilterPacket
from crownstone_core.packets.assetFilter.ExactMatchFilter import ExactMatchFilter
from crownstone_core.packets.assetFilter.FilterOutputPackets import FilterOutputDescriptionType, FilterOutputDescription
from crownstone_core.packets.assetFilter.InputDescriptionPackets import *
from crownstone_core.packets.assetFilter.FilterMetaDataPackets import FilterType, FilterMetaData, FilterFlags
from crownstone_core.packets.assetFilter.builders.AssetIdSourceBuilder import AssetIdSourceBuilder

_LOGGER = logging.getLogger(__name__)

class OptimizeOutputStrategy(Enum):
    # No optimization.
    NONE = 0,

    # Only crownstones that believe to be nearest. Note that this is not the same as nearest only, as it takes some time to settle.
    # Experimental: may be removed or changed in a later release.
    NEAREST = 1,

class AssetFilter(BasePacket):
    def __init__(self, filterId: int = None):
        """
        Class that helps to build an asset filter packet.
        1. Choose the filter input:        filterByX()
        2. Optionally, set configurations: setX()
        3. Choose the output:              outputX()

        :param filterId:     The ID of this filter, the index at which it is placed on the Crownstones.
        """
        self._filterId = filterId
        self._filterType: FilterType = None
        self._input: InputDescriptionPacket = None
        self._outputType: FilterOutputDescriptionType = None
        self._assetIdSourceBuilder: AssetIdSourceBuilder = None
        self._assets = []
        self._profileId = 255
        self._exclude = False
        self._maxFilterSize = 512

        # Cache
        self._packet: AssetFilterPacket = None
        self._crc = None

    def getCrc(self) -> int:
        self._buildIfNeeded()
        return self._crc

    def getFilterId(self) -> int:
        return self._filterId


    def filterByMacAddress(self, macAddresses: List[str]):
        """
        Assets are filtered by their MAC address.
        :param macAddresses: List of mac addresses to be added to the filter, in the form of "12:34:56:78:AB:CD".
        """
        self._resetCache()
        self._input = InputDescriptionMacAddress()
        self._assets = []
        for mac in macAddresses:
            self._assets.append(Conversion.address_to_uint8_array(mac))
        return self

    def filterByName(self, names: List[str], complete: bool = True):
        """
        Assets are filtered by their name (case sensitive).
        :param names:     List of names to be added filter.
        :param complete:  Whether to look for the complete or shortened name.
        """
        self._resetCache()
        adType = 0x09 if complete else 0x08
        self._input = InputDescriptionFullAdData(adType)

        self._assets = []
        for name in names:
            self._assets.append(Conversion.string_to_uint8_array(name))
        return self

    def filterByNameWithWildcards(self, name: str, complete: bool = True):
        """
        Assets are filtered by their name (case sensitive).
        :param name:      Name, with wildcards, to be added filter.
                          '?' matches any single character
                          '*' matches zero or more characters, only at the end of a name.
                          Example: "??_dev*" will match:  "my_dev", and "01_device"
                                             won't match: "your_device", or "_dev"
        :param complete:  Whether to look for the complete or shortened name.
        """
        self._resetCache()
        if len(name) > 31:
            raise CrownstoneException(CrownstoneError.INVALID_SIZE, f"Name is too long: {name}")
        bitmask = 0
        asset_name = ""
        for i in range(0, len(name)):
            if name[i] != '?':
                bitmask = set_bit(bitmask, i)
                asset_name += name[i]
        if name[-1] == '*':
            bitmask = set_bit(bitmask, len(name) - 1, False)
            asset_name = asset_name[:-1]
        else:
            # Set all remaining bits. If there is more data, it will be used as input.
            for i in range(len(name), 32):
                bitmask = set_bit(bitmask, i)

        _LOGGER.info(f"name={name} bitmask={bitmask:032b} asset_name={asset_name}")

        adType = 0x09 if complete else 0x08
        self._input = InputDescriptionMaskedAdData(adType, bitmask)
        self._assets = [Conversion.string_to_uint8_array(asset_name)]
        return self

    def filterByCompanyId(self, companyIds: List[int]):
        """
        Assets are filtered by their 16 bit company ID.
        :param companyIds: A list of 16 bit company IDs. As can be found on
                           https://www.bluetooth.com/specifications/assigned-numbers/company-identifiers/
        """
        self._resetCache()
        self._input = InputDescriptionMaskedAdData(0xFF, get_bitmask([0, 1]))
        self._assets = []
        for companyId in companyIds:
            self._assets.append(Conversion.uint16_to_uint8_array(companyId))
        return self

    def filterByAdData(self, adType: int, assets: List[list], bitmask: int = None):
        """
        Assets are recognized by the data of a single AD structure in a BLE advertisement.
        :param adType:  The 8 bit GAP number.
                        See "Generic Access Profile" on https://www.bluetooth.com/specifications/assigned-numbers/
        :param assets:  A list of assets, where each asset is represented by a list of bytes.
        :param bitmask: A 32 bits mask where the Nth bit represents the Nth byte in the AD data.
                        The data that is used as input for the filter, is a concatenation of all bytes that have their
                        associated bit set.
                        Example: if the AD data is: [10, 11, 12, 13, 14] and the bitmask is 22 (0000...00010110), then
                        the data that the filter uses is [11, 12, 14], and matched against each given asset.
        """
        self._resetCache()
        if bitmask is None:
            self._input = InputDescriptionFullAdData(adType)
        else:
            self._input = InputDescriptionMaskedAdData(adType, bitmask)
        self._assets = assets
        return self


    def setFilterId(self, filterId: int):
        """
        Set the ID of this filter, the index at which it is placed on the Crownstones.

        :param filterId:     The ID of this filter.
        """
        self._resetCache()
        self._filterId = filterId
        return self

    def setFilterType(self, filterType: FilterType):
        """
        Set the filter type.

        Each filter type has its pros and cons:
        - Cuckoo: stores a list of 2 bytes hashes of the asset data. This means there can be false positives.
            Use this type when:
                - You have many asset entries, each more than 2B of data.
                - You have asset entries of varying length.
            Examples:
                - 50 MAC addresses.
                - 10 names of different length.
        - Exact match: stores a list of the whole asset data. It only accepts same length entries, and does not compress the data.
            Use this type when:
                - You have a few assets to add.
                - You do not want any false positives.
            Examples:
                - 10 MAC addresses.
                - A list of company IDs.
                - A name.
        """
        self._resetCache()
        self._filterType = filterType
        return self

    def setExclude(self, exclude=True):
        """
        Make this an exclude filter.

        Any asset that passes an exclude filter, will be prevented from passing any other filter.
        An exclude filter has no output, so you don't have to choose one.

        :param exclude: True to make this an exclude filter.
        """
        self._resetCache()
        self._exclude = exclude
        return self

    def setProfileId(self, profileId: int):
        """
        By setting a profile ID, any asset advertisement that passes this filter will be treated as this profile ID
        for behaviours. If the localization cannot determine which room the asset is in, it will be still be treated as
        being in the sphere.

        :param profileId:    The profile ID for behaviour. 255 for no profile ID.
        """
        self._resetCache()
        self._profileId = profileId
        return self

    def setMaxFilterSize(self, maxFilterSize):
        """
        Set the max size (in bytes) of this filter.

        This will be taken into account when choosing the filter type.
        """
        self._resetCache()
        self._maxFilterSize = maxFilterSize
        return self

    def outputMacRssiReport(self):
        """
        If an asset advertisement passes the filter, the Crownstone will send a report to the hub with the assets' MAC address and the RSSI.
        """
        self._resetCache()
        self._outputType = FilterOutputDescriptionType.MAC_ADDRESS
        self._assetIdSourceBuilder = None
        return self

    def outputAssetId(self, basedOn: AssetIdSourceBuilder = None, optimizeStrategy: OptimizeOutputStrategy = OptimizeOutputStrategy.NONE) -> AssetIdSourceBuilder:
        """
        If an asset advertisement passes the filter, the Crownstones will attempt to localize it, and will identify it
        by a 3 byte asset ID. The asset ID is a hash over data from the advertisement, which can be different data than
        what it's filtered by. Select this data with the AssetIdBuilder.

        :param basedOn: Determines what data to base the short asset ID on.
        :param optimizeStrategy: Strategy to optimize number of mesh messages. Experimental, may be removed or changed in a later release.
        """
        self._resetCache()
        if optimizeStrategy == OptimizeOutputStrategy.NEAREST:
            self._outputType = FilterOutputDescriptionType.ASSET_ID_NEAREST_CROWNSTONE
        else:
            self._outputType = FilterOutputDescriptionType.ASSET_ID

        if basedOn is None:
            self._assetIdSourceBuilder = AssetIdSourceBuilder()
        else:
            self._assetIdSourceBuilder = basedOn
        return self._assetIdSourceBuilder

    def outputNone(self):
        """
        If an asset advertisement passes the filter, the Crownstone will only update its local presence information. No mesh traffic is generated.
        """
        self._resetCache()
        self._outputType = FilterOutputDescriptionType.NONE
        self._assetIdSourceBuilder = None
        return self


    def build(self) -> AssetFilterPacket:
        # Check variables
        if self._filterId is None:
            raise CrownstoneException(CrownstoneError.DATA_MISSING, f"No filter ID set.")
        if self._input is None:
            raise CrownstoneException(CrownstoneError.DATA_MISSING, f"No filter input set.")
        if (self._outputType is None) and (not self._exclude):
            raise CrownstoneException(CrownstoneError.DATA_MISSING, f"No filter output set.")

        # Determine output description
        outputType = FilterOutputDescriptionType.MAC_ADDRESS
        if self._outputType is not None:
            outputType = self._outputType

        inFormat = None
        if self._assetIdSourceBuilder is not None:
            inFormat = self._assetIdSourceBuilder.build()

        output = FilterOutputDescription(outputType, inFormat)

        # Remove duplicates.
        uniqueAssets = []
        for asset in self._assets:
            if asset not in uniqueAssets:
                uniqueAssets.append(asset)
        self._assets = uniqueAssets

        # Determine filter type to use if it hasn't been set.
        if self._filterType is None:
            equalSize = True
            assetSize = len(self._assets[0])
            totalSize = 0
            for asset in self._assets:
                if len(asset) != assetSize:
                    equalSize = False
                totalSize += len(asset)
            _LOGGER.debug(f"equalSize={equalSize} totalSize={totalSize}")

            filterOverhead = 100 # TODO: this is a very rough estimate.
            if totalSize + filterOverhead < self._maxFilterSize and equalSize:
                self._filterType = FilterType.EXACT_MATCH
            else:
                self._filterType = FilterType.CUCKOO

        # Build the meta data.
        metaData = FilterMetaData(self._filterType, self._input, output, self._profileId, FilterFlags(exclude=self._exclude))

        # Construct and fill the filter.
        if self._filterType == FilterType.EXACT_MATCH:
            filterData = ExactMatchFilter()
            for asset in self._assets:
                filterData.add(asset)
        elif self._filterType == FilterType.CUCKOO:
            # TODO: move this to cuckoo filter implementation.
            initialNestsPerBucket = 4
            requiredBucketCount = len(self._assets) / 0.95 / initialNestsPerBucket
            bucketCountLog2 = max(0, math.ceil(math.log2(requiredBucketCount)))
            bucketCount = math.pow(2, bucketCountLog2)
            nestsPerBucket = math.ceil(len(self._assets) / bucketCount)

            cuckooFilter = CuckooFilter(bucketCountLog2, nestsPerBucket)
            for asset in self._assets:
                if not cuckooFilter.add(asset):
                    raise CrownstoneException(CrownstoneError.INVALID_SIZE, "Failed to add asset to cuckoo filter.")
            filterData = cuckooFilter.getData()
        else:
            raise CrownstoneException(CrownstoneError.UNKNOWN_TYPE, f"Unknown filter type: {self._filterType}")

        self._packet = AssetFilterPacket(metaData, filterData)
        self._crc = crc32(self.serialize())
        return self._packet

    def _resetCache(self):
        if self._packet is not None:
            _LOGGER.debug("Removing cache")
            self._packet = None
            self._crc = None

    def _buildIfNeeded(self):
        if self._packet is None:
            self.build()

    def _serialize(self, writer: BufferWriter):
        self._buildIfNeeded()
        self._packet.serialize(writer)

    def __str__(self):
        self._buildIfNeeded()
        return f"AssetFilter(filterId={self._filterId} packet={self._packet})"




"""
 We will assume that there is a dictionary format which is used to provide the system with the behaviour data. We will raise an error on incorrect format. Otherwise we will parse it and return the Behaviour object.
 Expected format:
 {
    type: "BEHAVIOUR" | "TWILIGHT",
    data: {
        action: { data: number },
        time: TIME_PARSER_INPUT,
        presence: PRESENCE_PARSER_INPUT
        endCondition: ENDCONDITION_PARSER_INPUT
    },
    activeDays: {
      Mon: boolean,
      Tue: boolean,
      Wed: boolean,
      Thu: boolean,
      Fri: boolean,
      Sat: boolean,
      Sun: boolean
    },
    idOnCrownstone: number,
    profileIndex: number
 }
 """
import json

from crownstone_core.Exceptions import CrownstoneError, CrownstoneException
from crownstone_core.packets.behaviour.ActiveDays import ActiveDays
from crownstone_core.packets.behaviour.BehaviourBase import BehaviourBase
from crownstone_core.packets.behaviour.BehaviourTypes import BehaviourType, DAY_START_TIME_SECONDS_SINCE_MIDNIGHT
from crownstone_core.packets.behaviour.PresenceDescription import BehaviourPresenceType, BehaviourPresence
from crownstone_core.packets.behaviour.TimeDescription import BehaviourTimeContainer, BehaviourTime, BehaviourTimeType


def getFromDict(dict, key):
    val = None
    if key in dict: val = dict[key]
    return val

def BehaviourDictionaryParser(dict, dayStartTimeSecondsSinceMidnight = DAY_START_TIME_SECONDS_SINCE_MIDNIGHT):
    # optional variables
    profileIndex   = getFromDict(dict, "profileIndex")
    typeString     = getFromDict(dict, "type")
    data           = getFromDict(dict, "data")
    activeDays     = getFromDict(dict, "activeDays")
    idOnCrownstone = getFromDict(dict, "idOnCrownstone")
    
    if profileIndex  is None: raise CrownstoneException(CrownstoneError.PROFILE_INDEX_MISSING)
    if typeString    is None: raise CrownstoneException(CrownstoneError.TYPE_MISSING)
    if data          is None: raise CrownstoneException(CrownstoneError.DATA_MISSING)
    if activeDays    is None: raise CrownstoneException(CrownstoneError.ACTIVE_DAYS_MISSING)
    
    behaviourType = BehaviourType.behaviour
    if   typeString == "BEHAVIOUR": behaviourType = BehaviourType.behaviour
    elif typeString == "TWILIGHT":  behaviourType = BehaviourType.twilight

    if type(data) == str:
        data = json.loads(data)

    # optional variables
    actionDict   = getFromDict(data, "action")
    timeDict     = getFromDict(data, "time")
    presence     = getFromDict(data, "presence")
    endCondition = getFromDict(data, "endCondition")

    if actionDict is None: raise CrownstoneException(CrownstoneError.BEHAVIOUR_ACTION_MISSING)
    if timeDict   is None: raise CrownstoneException(CrownstoneError.BEHAVIOUR_TIME_MISSING)

  
    intensity = getFromDict(actionDict, "data")
    
    if intensity is None: raise CrownstoneException(CrownstoneError.BEHAVIOUR_INTENSITY_MISSING)
    
    activeDayObject = ActiveDayParser(activeDays)
    timeObject      = TimeParser(timeDict, dayStartTimeSecondsSinceMidnight)
    
    behaviour = BehaviourBase(
        profileIndex  = profileIndex,
        behaviourType = behaviourType,
        intensity     = intensity,
        activeDays    = activeDayObject,
        time          = timeObject
    )
    
    if idOnCrownstone is not None:
        behaviour.idOnCrownstone = idOnCrownstone
        
    if presence is not None:
        if behaviourType == BehaviourType.twilight:
            raise CrownstoneException(CrownstoneError.TWILIGHT_CANT_HAVE_PRESENCE)
        
        presenceObject = PresenceParser(presence)
        behaviour.presence = presenceObject
    
    
    if endCondition is not None:
        if behaviourType == BehaviourType.twilight:
            raise CrownstoneException(CrownstoneError.TWILIGHT_CANT_HAVE_END_CONDITION)
        
        endConditionObject = EndConditionParser(endCondition)
        behaviour.endCondition = endConditionObject
        behaviour.type = BehaviourType.smartTimer
    
    
    return behaviour


"""
 There are a few possible formats. We will parse and validate
"""
def ActiveDayParser(dict):
    Monday    = getFromDict(dict, "Mon")
    Tuesday   = getFromDict(dict, "Tue")
    Wednesday = getFromDict(dict, "Wed")
    Thursday  = getFromDict(dict, "Thu")
    Friday    = getFromDict(dict, "Fri")
    Saturday  = getFromDict(dict, "Sat")
    Sunday    = getFromDict(dict, "Sun")
    
    activeDays = ActiveDays()
    
    if Monday is not None and Tuesday is not None and Wednesday is not None and Thursday is not None and Friday is not None and Saturday is not None and Sunday is not None:
        activeDays.Monday    = Monday
        activeDays.Tuesday   = Tuesday
        activeDays.Wednesday = Wednesday
        activeDays.Thursday  = Thursday
        activeDays.Friday    = Friday
        activeDays.Saturday  = Saturday
        activeDays.Sunday    = Sunday
        
        if activeDays.getMask() == 0:
            raise CrownstoneException(CrownstoneError.NO_ACTIVE_DAYS)
        
        return activeDays
    else:
        raise CrownstoneException(CrownstoneError.ACTIVE_DAYS_INVALID)
    


"""
 There are a few possible formats. We will parse and validate
"""
def TimeParser(dict, dayStartTimeSecondsSinceMidnight):
    type = getFromDict(dict, "type")

    if type is None: raise CrownstoneException(CrownstoneError.NO_TIME_TYPE)

    if type == "ALL_DAY":
        return BehaviourTimeContainer(
            BehaviourTime().fromType(BehaviourTimeType.afterMidnight, dayStartTimeSecondsSinceMidnight),
            BehaviourTime().fromType(BehaviourTimeType.afterMidnight, dayStartTimeSecondsSinceMidnight)
        )
    
    elif type == "RANGE":
        fromTime = getFromDict(dict, "from")
        toTime   = getFromDict(dict, "to")

        if fromTime is None: raise CrownstoneException(CrownstoneError.MISSING_FROM_TIME)
        if toTime   is None: raise CrownstoneException(CrownstoneError.MISSING_TO_TIME)

        fromTimeType = getFromDict(fromTime, "type")
        toTimeType   = getFromDict(toTime,   "type")

        if fromTimeType is None: raise CrownstoneException(CrownstoneError.MISSING_FROM_TIME_TYPE)
        if toTimeType   is None: raise CrownstoneException(CrownstoneError.MISSING_TO_TIME_TYPE)

        fromResult = None
        toResult   = None
        
        if fromTimeType == "CLOCK":
            fromData = getFromDict(fromTime, "data")
        
            if fromData is None: raise CrownstoneException(CrownstoneError.MISSING_FROM_TIME_DATA)

            hours   = getFromDict(fromData, "hours")
            minutes = getFromDict(fromData, "minutes")

            if hours   is None: raise CrownstoneException(CrownstoneError.MISSING_FROM_TIME_DATA)
            if minutes is None: raise CrownstoneException(CrownstoneError.MISSING_FROM_TIME_DATA)

            fromResult = BehaviourTime().fromTime(hours, minutes)

        elif fromTimeType == "SUNSET":
            offsetSeconds = 0
            offsetMinutes = getFromDict(fromTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60*offsetMinutes

            fromResult = BehaviourTime().fromType(BehaviourTimeType.afterSunset, offsetSeconds)
        
        elif fromTimeType == "SUNRISE":
            offsetSeconds = 0
            offsetMinutes = getFromDict(fromTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60 * offsetMinutes
            
            fromResult = BehaviourTime().fromType(BehaviourTimeType.afterSunrise, offsetSeconds)
        
        else:
            raise CrownstoneException(CrownstoneError.INVALID_TIME_FROM_TYPE)
        
        
        
        if toTimeType == "CLOCK":
            toData = getFromDict(toTime, "data")

            if toData is None: raise CrownstoneException(CrownstoneError.MISSING_TO_TIME_DATA)

            hours = getFromDict(toData, "hours")
            minutes = getFromDict(toData, "minutes")

            if hours is None: raise CrownstoneException(CrownstoneError.INVALID_TO_DATA)
            if minutes is None: raise CrownstoneException(CrownstoneError.INVALID_TO_DATA)

            toResult = BehaviourTime().fromTime(hours, minutes)

        elif toTimeType == "SUNSET":
            offsetSeconds = 0
            offsetMinutes = getFromDict(toTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60 * offsetMinutes

            toResult = BehaviourTime().fromType(BehaviourTimeType.afterSunset, offsetSeconds)

        elif toTimeType == "SUNRISE":
            offsetSeconds = 0
            offsetMinutes = getFromDict(toTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60 * offsetMinutes

            toResult = BehaviourTime().fromType(BehaviourTimeType.afterSunrise, offsetSeconds)
        
        else:
            raise CrownstoneException(CrownstoneError.INVALID_TIME_TO_TYPE)
        
        return BehaviourTimeContainer(fromResult, toResult)
    else:
        raise CrownstoneException(CrownstoneError.INVALID_TIME_TYPE)
    
    


"""
 There are a few possible formats. We will parse and validate
"""
def PresenceParser(dict):
    type = getFromDict(dict, "type")
    if type is None: raise CrownstoneException(CrownstoneError.NO_PRESENCE_TYPE)

    if type == "IGNORE":
        return BehaviourPresence()
    elif type == "SOMEBODY" or type == "NOBODY":
        data = getFromDict(dict, "data")
        if data is None: raise CrownstoneException(CrownstoneError.NO_PRESENCE_DATA)

        delay = getFromDict(dict, "delay")
        if delay is None: raise CrownstoneException(CrownstoneError.NO_PRESENCE_DELAY)

        dataType = getFromDict(data, "type")
        if dataType is None: raise CrownstoneException(CrownstoneError.NO_PRESENCE_TYPE)

        if dataType == "SPHERE":
            if type == "SOMEBODY":
                return BehaviourPresence().setSpherePresence(BehaviourPresenceType.someoneInSphere, delayInSeconds = delay)
            else:
                return BehaviourPresence().setSpherePresence(BehaviourPresenceType.nobodyInSphere, delayInSeconds = delay)

        elif dataType == "LOCATION":
            locationIdArray = getFromDict(data, "locationIds")
            if locationIdArray is None: raise CrownstoneException(CrownstoneError.NO_PRESENCE_LOCATION_IDS)

            if type == "SOMEBODY":
                return BehaviourPresence().setLocationPresence(BehaviourPresenceType.somoneInLocation, locationIds=locationIdArray, delayInSeconds=delay)
            
            else:
                return BehaviourPresence().setLocationPresence(BehaviourPresenceType.nobodyInLocation, locationIds=locationIdArray, delayInSeconds=delay)
        else:
            raise CrownstoneException(CrownstoneError.NO_PRESENCE_DATA)
    else:
        raise CrownstoneException(CrownstoneError.INVALID_PRESENCE_TYPE)
    
    
    


"""
 There are a few possible formats. We will parse and validate
"""
def EndConditionParser(dict):
    type     = getFromDict(dict, "type")
    presence = getFromDict(dict, "presence")
    if type     is None: raise CrownstoneException(CrownstoneError.NO_END_CONDITION_TYPE)
    if presence is None: raise CrownstoneException(CrownstoneError.NO_END_CONDITION_PRESENCE)

    presenceObject = PresenceParser(presence)
    
    return presenceObject




from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvErrorPacket:
    def __init__(self):
        self.type = AdvType.CROWNSTONE_ERROR

        self.crownstoneId              = None
        self.errorsBitmask             = None
        self.errorTimestamp            = None
        self.flags                     = None
        self.temperature               = None
        self.timestamp                 = None
        self.uniqueIdentifier          = None
        self.powerUsageReal            = None

    def __str__(self):
        return f"{self.type}\n"\
           f"crownstoneId:      {self.crownstoneId    }\n" \
           f"errorsBitmask:     {self.errorsBitmask   }\n" \
           f"errorTimestamp:    {self.errorTimestamp  }\n" \
           f"flags:             \n{self.flags           }\n" \
           f"temperature:       {self.temperature     }\n" \
           f"timestamp:         {self.timestamp       }\n" \
           f"uniqueIdentifier:  {self.uniqueIdentifier}\n" \
           f"powerUsageReal:    {self.powerUsageReal  }\n"



import math

from crownstone_core.packets.assetFilter.FilterCommandPackets import UploadFilterChunkPacket
from crownstone_core.packets.assetFilter.builders.AssetFilter import AssetFilter


class FilterChunker:

    def __init__(self, assetFilter: AssetFilter, maxChunkSize=128):
        self.filterId = assetFilter.getFilterId()
        self.filterBuffer = assetFilter.serialize()

        self.index = 0
        self.maxChunkSize = maxChunkSize

    def getAmountOfChunks(self) -> int:
        totalSize = len(self.filterBuffer)
        count = math.floor(totalSize/self.maxChunkSize)
        if totalSize % self.maxChunkSize > 0:
            count += 1
        return count

    def getChunk(self) -> [int]:
        totalSize = len(self.filterBuffer)
        if totalSize > self.maxChunkSize:
            chunkSize = min(self.maxChunkSize, totalSize - self.index)
            chunkData = self.filterBuffer[self.index : self.index + chunkSize]
            cmd = UploadFilterChunkPacket(self.filterId, self.index, totalSize, chunkSize, chunkData)
            self.index += self.maxChunkSize
            return cmd.serialize()
        else:
            return UploadFilterChunkPacket(self.filterId, self.index, totalSize, totalSize, self.filterBuffer).serialize()



from enum import IntEnum

from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.BufferReader import BufferReader

BOOTLOADER_INFO_PROTOCOL = 1

class BootloaderBuildType(IntEnum):
    UNKNOWN                 = 0,
    DEBUG                   = 1,
    RELEASE                 = 2,
    RELEASE_WITH_DEBUG      = 3,
    RELEASE_MIN_SIZE        = 4,

class BootloaderInfoPacket(BasePacket):
    def __init__(self, data = None):
        self.protocol = 0
        self.dfuVersion = 0
        self.major = 0
        self.minor = 0
        self.patch = 0
        self.preReleaseVersion = 0
        self.buildType = BootloaderBuildType.UNKNOWN

        if data is not None:
            self.deserialize(data)

    def _deserialize(self, reader: BufferReader):
        self.protocol = reader.getUInt8()
        self.dfuVersion = reader.getUInt16()
        self.major = reader.getUInt8()
        self.minor = reader.getUInt8()
        self.patch = reader.getUInt8()
        self.preReleaseVersion = reader.getUInt8()
        self.buildType = BootloaderBuildType(reader.getUInt8())

    def toString(self):
        return f"BootloaderInfoPacket(" \
               f"protocol={self.protocol}, " \
               f"dfuVersion={self.dfuVersion}, " \
               f"major={self.major}, " \
               f"minor={self.minor}, " \
               f"patch={self.patch}, " \
               f"preReleaseVersion={self.preReleaseVersion}, " \
               f"buildType={self.buildType})"

    def __str__(self):
        return self.toString()



from typing import List

from crownstone_core.Exceptions import CrownstoneException, CrownstoneError
from crownstone_core.packets.assetFilter.FilterCommandPackets import FilterSummaryPacket
from crownstone_core.packets.assetFilter.builders.AssetFilter import AssetFilter
from crownstone_core.util.BufferWriter import BufferWriter
from crownstone_core.util.CRC import crc32

def get_master_crc_from_filters(filters: List[AssetFilter], filterSummaries: List[FilterSummaryPacket] = None) -> int:
    """
    Get the master CRC from a list of filters.

    :param filters:          A list of asset filters with filter ID, that are uploaded to the Crowstone.
    :param filterSummaries:  A list of filter summaries that are already on the Crownstone.
    :returns:                The master CRC.
    """
    input_data = []
    for assetFilter in filters:
        filterId = assetFilter.getFilterId()
        crc = assetFilter.getCrc()
        input_data.append([filterId, crc])

    if filterSummaries is None:
        filterSummaries = []

    for filterSummary in filterSummaries:
        filterId = filterSummary.id
        crc = filterSummary.crc
        input_data.append([filterId, crc])

    return get_master_crc_from_filter_crcs(input_data)

def get_master_crc_from_filter_crcs(input_data : [[int, int]]) -> int:
    """
    Get the master CRC from filter CRCs.

    :param input_data:  A list of [filterId, filterCRC].
    :returns:           The master CRC.
    """
    def sort_method(val):
        return val[0]

    input_data.sort(key=sort_method)

    # Check for duplicate filter IDs.
    for i in range(0, len(input_data) - 1):
        if input_data[i][0] == input_data[i+1][0]:
            raise CrownstoneException(CrownstoneError.INVALID_INPUT, "Cannot have 2 filters with the same ID.")

    writer = BufferWriter()
    for id_and_filter_crc in input_data:
        writer.putUInt8(id_and_filter_crc[0])
        writer.putUInt32(id_and_filter_crc[1])

    return crc32(writer.getBuffer())



from crownstone_core.Exceptions import CrownstoneException, CrownstoneError

from crownstone_core.util.BufferWriter import BufferWriter

from crownstone_core.packets.assetFilter.FilterDataPackets import FilterData


class ExactMatchFilter(FilterData):
    def __init__(self):
        self.itemCount = 0
        self.itemSize = 0
        self.items = []

    def add(self, item: list):
        """
        Add an item to the filter.

        :param item: Byte array representation of the item.
        """
        if item in self.items:
            # Avoid duplicates
            return

        if self.itemCount:
            if self.itemSize != len(item):
                raise CrownstoneException(CrownstoneError.INVALID_SIZE, f"Must be same size as other items ({self.itemSize})")
        else:
            self.itemSize = len(item)
        self.items.append(item)
        self.itemCount = len(self.items)

    def _serialize(self, writer: BufferWriter):
        writer.putUInt8(self.itemCount)
        writer.putUInt8(self.itemSize)

        self.items.sort()
        for item in self.items:
            writer.putBytes(item)

    def __str__(self):
        return f"ExactMatchFilter(" \
               f"itemCount={self.itemCount} " \
               f"itemSize={self.itemSize} " \
               f"items={self.items})"


from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvHubState:
    def __init__(self):
        self.type = AdvType.HUB_STATE

        self.crownstoneId     = None
        self.hubFlags         = None
        self.hubData          = None
        self.timestamp        = None
        self.uniqueIdentifier = None
        self.validation       = None

    def __str__(self):
        return f"{self.type}\n"\
           f"crownstoneId     {self.crownstoneId    }\n" \
           f"hubFlags         {self.hubFlags        }\n" \
           f"hubData          {self.hubData         }\n" \
           f"timestamp        {self.timestamp       }\n" \
           f"uniqueIdentifier {self.uniqueIdentifier}\n" \
           f"validation       {self.validation      }\n"



from enum import IntEnum

from crownstone_core.packets.assetFilter.FilterOutputPackets import FilterOutputDescription
from crownstone_core.packets.assetFilter.InputDescriptionPackets import *
from crownstone_core.util.BufferWriter import BufferWriter
from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.util.Bitmasks import set_bit


class FilterType(IntEnum):
    CUCKOO = 0
    EXACT_MATCH = 1



class FilterFlags(BasePacket):
    def __init__(self, exclude = False):
        self.exclude: bool = exclude
        self.bitmask: int  = 0
        self._calc_bitmask()

    def _calc_bitmask(self):
        self.bitmask = set_bit(self.bitmask, 0, self.exclude)

    def _serialize(self, writer: BufferWriter):
        self._calc_bitmask()
        writer.putUInt8(self.bitmask)

    def __str__(self):
        return f"FilterFlags(" \
               f"exclude={self.exclude})"



class FilterMetaData(BasePacket):
    """
    Common metadata of an asset filter.
    """
    def __init__(self,
                 type: FilterType,
                 filterInput: InputDescriptionMacAddress or InputDescriptionFullAdData or InputDescriptionMaskedAdData,
                 filterOutput: FilterOutputDescription,
                 profileId: int = 255,
                 flags: FilterFlags = FilterFlags()):
        self.type              = type
        self.flags             = flags
        self.profileId         = profileId
        self.filterInput       = filterInput
        self.filterOutput      = filterOutput

    def _serialize(self, writer: BufferWriter):
        writer.putUInt8(self.type)
        self.flags.serialize(writer)
        writer.putUInt8(self.profileId)
        self.filterInput.serialize(writer)
        self.filterOutput.serialize(writer)

    def __str__(self):
        return f"FilterMetaData(" \
               f"type={self.type} " \
               f"flags={self.flags} "\
               f"profileId={self.profileId} " \
               f"filterInput={self.filterInput} " \
               f"filterOutput={self.filterOutput})"



from crownstone_core.packets.serviceDataParsers.containers.elements.AdvCrownstoneErrorBitmask import AdvCrownstoneErrorBitmask

from crownstone_core.protocol.SwitchState import SwitchState
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.packets.serviceDataParsers.containers.AdvCrownstoneSetupState import AdvCrownstoneSetupState
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvFlags import AdvFlags


def parseSetupState(reader: BufferReader):
    packet = AdvCrownstoneSetupState()

    packet.switchState  = SwitchState(reader.getUInt8())
    packet.flags        = AdvFlags(reader.getUInt8())
    packet.temperature  = reader.getInt8()
    powerFactor         = reader.getInt8()
    realPower           = reader.getInt16()

    packet.powerFactor = float(powerFactor) / 127.0

    # we cannot have a 0 for a power factor.To avoid division by 0, we set it to be either 0.01 or -0.01
    if 0 <= packet.powerFactor < 0.01:
        packet.powerFactor = 0.01
    elif -0.01 < packet.powerFactor < 0:
        packet.powerFactor = -0.01

    packet.powerUsageReal     = float(realPower) / 8.0
    packet.powerUsageApparent = packet.powerUsageReal / packet.powerFactor
    packet.errorsBitmask      = AdvCrownstoneErrorBitmask(reader.getUInt32())
    packet.uniqueIdentifier   = reader.getUInt8()

    return packet


from crownstone_core.packets.assetFilter.FilterCommandPackets import *
from crownstone_core.protocol.BlePackets import ControlPacket, FactoryResetPacket
from crownstone_core.protocol.BluenetTypes import ControlType
from crownstone_core.util.BufferWriter import BufferWriter
from crownstone_core.util.Conversion import Conversion

"""
TODO: rename functions to getXxxArray or so, as they output a serialized array, not a typed packet.
"""

class ControlPacketsGenerator:

    @staticmethod
    def getFactoryResetPacket():
        return Conversion.uint32_to_uint8_array(0xdeadbeef)


    @staticmethod
    def getCommandFactoryResetPacket():
        return FactoryResetPacket().serialize()

    @staticmethod
    def getSwitchCommandPacket(switchVal: int):
        """
        :param switchVal: percentage [0..100] or special value (SwitchValSpecial).
        """
        return ControlPacket(ControlType.SWITCH).loadUInt8(switchVal).serialize()

    @staticmethod
    def getResetPacket():
        return ControlPacket(ControlType.RESET).serialize()

    @staticmethod
    def getPutInDFUPacket():
        return ControlPacket(ControlType.GOTO_DFU).serialize()

    @staticmethod
    def getDisconnectPacket():
        return ControlPacket(ControlType.DISCONNECT).serialize()

    @staticmethod
    def getRelaySwitchPacket(turnOn: bool):
        """
        :param turnOn: True to turn relay on.
        """
        switchVal = 0
        if turnOn:
            switchVal = 1

        return ControlPacket(ControlType.RELAY).loadUInt8(switchVal).serialize()

    @staticmethod
    def getDimmerSwitchPacket(intensity: int):
        """
        :param intensity: percentage [0..100]
        """
        return ControlPacket(ControlType.PWM).loadUInt8(intensity).serialize()


    @staticmethod
    def getResetErrorPacket(errorMask):
        return ControlPacket(ControlType.RESET_ERRORS).loadUInt32(errorMask).serialize()

    @staticmethod
    def getSetTimePacket(time):
        """
        This is a LOCAL timestamp since epoch in seconds

        so if you live in GMT + 1 add 3600 to the timestamp
        :param time:
        :return:
        """
        return ControlPacket(ControlType.SET_TIME).loadUInt32(time).serialize()

    @staticmethod
    def getAllowDimmingPacket(allow: bool):
        """

        :param allow: bool
        :return:
        """

        allowByte = 0
        if allow:
            allowByte = 1

        return ControlPacket(ControlType.ALLOW_DIMMING).loadUInt8(allowByte).serialize()

    @staticmethod
    def getLockSwitchPacket(lock: bool):
        """
        :param lock: bool
        :return:
        """

        lockByte = 0
        if lock:
            lockByte = 1

        return ControlPacket(ControlType.LOCK_SWITCH).loadUInt8(lockByte).serialize()

    @staticmethod
    def getSetupPacket(
        crownstoneId: int,
        sphereId: int,
        adminKey,
        memberKey,
        basicKey,
        serviceDataKey,
        localizationKey,
        meshDeviceKey,
        meshAppKey,
        meshNetworkKey,
        ibeaconUUID: str,
        ibeaconMajor: int,
        ibeaconMinor: int
    ):
        """
        :param crownstoneId:  		uint8 number
        :param sphereId:  	     	uint8 number
        :param adminKey:      		byteString (no conversion required)
        :param memberKey:     		byteString (no conversion required)
        :param basicKey:      		byteString (no conversion required)
        :param serviceDataKey: 	    byteString (no conversion required)
        :param localizationKey: 	byteString (no conversion required)
        :param meshDeviceKey: 	    byteString (no conversion required)
        :param meshAppKey: 	        byteString (no conversion required)
        :param meshNetworkKey: 	    byteString (no conversion required)
        :param ibeaconUUID: 		string  (ie. "1843423e-e175-4af0-a2e4-31e32f729a8a")
        :param ibeaconMajor:        uint16 number
        :param ibeaconMinor:        uint16 number
        :return:
        """
        data = []
        data.append(crownstoneId)
        data.append(sphereId)

        data += list(adminKey)
        data += list(memberKey)
        data += list(basicKey)
        data += list(serviceDataKey)
        data += list(localizationKey)

        MDKey = meshDeviceKey
        if type(meshDeviceKey) is str:
            MDKey = Conversion.ascii_or_hex_string_to_16_byte_array(meshDeviceKey)

        data += list(MDKey)
        data += list(meshAppKey)
        data += list(meshNetworkKey)

        data += Conversion.ibeaconUUIDString_to_reversed_uint8_array(ibeaconUUID)
        data += Conversion.uint16_to_uint8_array(ibeaconMajor)
        data += Conversion.uint16_to_uint8_array(ibeaconMinor)

        return ControlPacket(ControlType.SETUP).loadByteArray(data).serialize()


    @staticmethod
    def getIBeaconConfigIdPacket(id, timestamp, interval):
        data = []
        data.append(id)
        data += Conversion.uint32_to_uint8_array(timestamp)
        data += Conversion.uint16_to_uint8_array(interval)

        return ControlPacket(ControlType.SET_IBEACON_CONFIG_ID).loadByteArray(data).serialize()

    @staticmethod
    def getPowerSamplesRequestPacket(samplesType, index):
        buffer = BufferWriter()
        buffer.putUInt8(samplesType)
        buffer.putUInt8(index)
        data = buffer.getBuffer()
        return ControlPacket(ControlType.GET_POWER_SAMPLES).loadByteArray(data).serialize()


    @staticmethod
    def getRemoveFilterPacket(filterId: int) -> [int]:
        removecommand = RemoveFilterPacket(filterId)
        return ControlPacket(ControlType.ASSET_FILTER_REMOVE).loadByteArray(removecommand.serialize()).serialize()


    @staticmethod
    def getGetFilterSummariesPacket() -> [int]:
        return ControlPacket(ControlType.ASSET_FILTER_GET_SUMMARIES).serialize()

    @staticmethod
    def getCommitFilterChangesPacket(masterVersion: int, masterCrc: int) -> [int]:
        commitcommand = CommitFilterChangesPacket(masterVersion, masterCrc)
        return ControlPacket(ControlType.ASSET_FILTER_COMMIT_CHANGES).loadByteArray(commitcommand.serialize()).serialize()

    @staticmethod
    def getUploadFilterPacket(chunk: [int]) -> [int]:
        """
        TODO: type chunk param
        """
        return ControlPacket(ControlType.ASSET_FILTER_UPLOAD).loadByteArray(chunk).serialize()


CROWNSTONE_PLUG_ADVERTISEMENT_SERVICE_UUID = "C001"
CROWNSTONE_BUILTIN_ADVERTISEMENT_SERVICE_UUID = "C002"
CROWNSTONE_GUIDESTONE_ADVERTISEMENT_SERVICE_UUID  = "C003"
DFU_ADVERTISEMENT_SERVICE_UUID = "00001530-1212-efde-1523-785feabcd123"

class CSServices:
	DeviceInformation         = "0000180a-0000-1000-8000-00805f9b34fb"
	CrownstoneService         = "24f00000-7d10-4805-bfc1-7663a01c3bff"
	SetupService              = "24f10000-7d10-4805-bfc1-7663a01c3bff"

class DFUServices:
	DFU = DFU_ADVERTISEMENT_SERVICE_UUID




from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType


class AdvCrownstoneState:
    def __init__(self):
        self.type = AdvType.CROWNSTONE_STATE

        self.crownstoneId              = None
        self.switchState               = None
        self.flags                     = None
        self.globalFlags               = None
        self.temperature               = None
        self.powerFactor               = None
        self.powerUsageReal            = None
        self.powerUsageApparent        = None
        self.accumulatedEnergy         = None
        self.timestamp                 = None
        self.uniqueIdentifier          = None
        self.validation                = None

    def __str__(self):
        return f"{self.type}\n"\
           f"crownstoneId:         {self.crownstoneId      }\n" \
           f"switchState:          {self.switchState       }\n" \
           f"flags:                {self.flags             }\n" \
           f"globalFlags:          {self.globalFlags       }\n" \
           f"temperature:          {self.temperature       }\n" \
           f"powerFactor:          {self.powerFactor       }\n" \
           f"powerUsageReal:       {self.powerUsageReal    }\n" \
           f"powerUsageApparent:   {self.powerUsageApparent}\n" \
           f"accumulatedEnergy:    {self.accumulatedEnergy }\n" \
           f"timestamp:            {self.timestamp         }\n" \
           f"uniqueIdentifier:     {self.uniqueIdentifier  }\n" \
           f"validation:           {self.validation        }\n"


from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter


class MicroappMessagePacket(BasePacket):
	def __init__(self, header: MicroappHeaderPacket, payload: bytearray or list):
		self.header = header
		self.payload = payload

	@staticmethod
	def fromData(data: bytearray):
		instance = MicroappMessagePacket(MicroappHeaderPacket(0, 0), [])
		instance.deserialize(data)
		return instance

	def _deserialize(self, reader: BufferReader):
		self.header.deserialize(reader)
		self.payload = reader.getRemainingBytes()

	def _serialize(self, writer: BufferWriter):
		self.header.serialize(writer)
		writer.putBytes(self.payload)

	def __str__(self):
		return f"MicroappMessagePacket(" \
		       f"header={self.header} " \
		       f"payload={[self.payload]})"



from crownstone_core.packets.BasePacket import BasePacket
from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.util.BufferReader import BufferReader
from crownstone_core.util.BufferWriter import BufferWriter


class MicroappTestsPacket(BasePacket):
	def __init__(self, data=None):
		self.raw = []
		self.hasData = False
		self.checksum = 0
		self.enabled = False
		self.boot = 0
		self.memory = 0
		self.reserved = 0

		if data is not None:
			self.deserialize(data)

	def _deserialize(self, reader: BufferReader):
		data1 = reader.getUInt8()
		data2 = reader.getUInt8()
		self.raw = [data1, data2]

		# 1 bit hasData
		self.hasData = (data1 & 0x01) == 1
		data1 = data1 >> 1

		# 2 bits checksum
		self.checksum = data1 & 0x03
		data1 = data1 >> 2

		# 1 bit enabled
		self.enabled = (data1 & 0x01) == 1
		data1 = data1 >> 1

		# 2 bits boot
		self.boot = data1 & 0x03
		data1 = data1 >> 2

		# 1 bit memory
		self.memory = data1 & 0x01
		data1 = data1 >> 1

		# 9 bits reserved
		self.reserved = data1

	def __str__(self):
		return f"MicroappTestsPacket(" \
		       f"raw={self.raw}, " \
		       f"hasData={self.hasData}, " \
		       f"checksum={self.checksum}, " \
		       f"enabled={self.enabled}, "\
		       f"boot={self.boot}, " \
		       f"memory={self.memory}, " \
		       f"reserved={[self.reserved]})"



from crownstone_core.util.BufferReader import BufferReader

class RamStatsPacket:
	def __init__(self, data):
		self.minStackEnd = 0
		self.maxHeapEnd = 0
		self.minFree = 0
		self.numSbrkFails = 0
		self.load(data)

	def load(self, data):
		"""
		Parses data buffer to set member variables.

		data : list of bytes

		Raises exception when parsing fails.
		"""
		streamBuf = BufferReader(data)
		self.minStackEnd = streamBuf.getUInt32()
		self.maxHeapEnd = streamBuf.getUInt32()
		self.minFree = streamBuf.getUInt32()
		self.numSbrkFails = streamBuf.getUInt32()

	@staticmethod
	def size():
		return 4 + 4 + 4 + 4

	def toString(self):
		return f"RamStatsPacket(minStackEnd=0x{self.minStackEnd:08X} maxHeapEnd=0x{self.maxHeapEnd:08X} minFree={self.minFree} numSbrkFails={self.numSbrkFails})"

	def __str__(self):
		return self.toString()